#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError, PublicationEmbargoWarning

class Hnscc(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the hnscc dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["0.1", "2.0"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "0.1": [
                "HNSCC.strelka.sorted.filtered.annovar.hg19_multianno_filtered.maf.txt.gz",
                "Proteomics_DIA_Gene_level_Normal.cct.gz",
                "Proteomics_DIA_Gene_level_Tumor.cct.gz",
                "RNAseq_RSEM_UQ_log2.cct.gz",
                "RNAseq_circ_RSEM_UQ_log2.cct.gz",
                "SCNA_gene_level.cct.gz",
                "clinic.tsi.gz"],
            "2.0": [
                "circRNAseq_RSEM_UQ_log2_Combined.cct.gz",
                "HN_followUp_9_24.xlsx",
                "Meta_table.tsv.gz",
                "microRNA_log2_Combined.cct.gz",
                "Phosphoproteomics_TMT_site_level_combined_all.cct.gz",
                "Proteomics_TMT_gene_level_combined_all.cct.gz",
                "RNAseq_RSEM_UQ_Combined.cct.gz",
                "SCNA_log2_gene_level.cct.gz",
                "SomaticMutations_maf.tsv.gz"],
        }

        super().__init__(cancer_type="hnscc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Our dataframe name will be the first section of file name (i.e. proteomics.txt.gz becomes proteomics)

            if file_name == "SCNA_gene_level.cct.gz" or file_name == "SCNA_log2_gene_level.cct.gz":
                df = pd.read_csv(file_path, sep="\t")

                if self._version == "2.0":
                    df = df.set_index('gene_symbol')

                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.columns.name=None
                df.index.name = "Patient_ID"
                self._data["CNV"] = df

            elif file_name == "microRNA_log2_Combined.cct.gz" and self._version == "2.0":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.sort_index()
                df = df.transpose()

                # Reformat patient ids
                df.index = df.index.str.replace(r'-T$', '', 1, regex=True)
                df.index = df.index.str.replace(r'-N$', '.N', 1, regex=True)

                self._data["miRNA"] = df

            elif file_name == "RNAseq_RSEM_UQ_log2.cct.gz" or file_name == "RNAseq_RSEM_UQ_Combined.cct.gz":
                df = pd.read_csv(file_path, sep="\t")

                if self._version == "2.0":
                    df = df.set_index('Idx')

                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.columns.name=None

                if self._version == "0.1":
                    df.index = df.index.str.replace(r'\.', '-', 1, regex=True)
                    df.index = df.index.str.replace(r'\.T$', '', 1, regex=True)
                elif self._version == "2.0":
                        df.index = df.index.str.replace(r'-T$', '', 1, regex=True)
                        df.index = df.index.str.replace(r'-N$', '.N', 1, regex=True)

                df.index.name = "Patient_ID"
                self._data["transcriptomics"] = df

            elif file_name == "RNAseq_circ_RSEM_UQ_log2.cct.gz" or file_name == "circRNAseq_RSEM_UQ_log2_Combined.cct.gz":
                df = pd.read_csv(file_path, sep='\t')
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.columns.name=None

                if self._version == "0.1":
                    df.index = df.index.str.replace(r'\.', '-', 1, regex=True) # We want all the patientIDs to have the the format C3L-00977, and these have the form C3L.00977.N, so we need to replace the first "." with a "-"
                    df.index = df.index.str.replace(r'\.T$', '', 1, regex=True)

                elif self._version == "2.0":
                    df.index = df.index.str.replace(r'-T$', '', 1, regex=True)
                    df.index = df.index.str.replace(r'-N$', '.N', 1, regex=True)

                df.index.name = "Patient_ID"
                self._data["circular_RNA"] = df

            elif file_name == "HNSCC.strelka.sorted.filtered.annovar.hg19_multianno_filtered.maf.txt.gz" or file_name == "SomaticMutations_maf.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")

                if self._version == "0.1":
                    df = df.rename(columns={"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol_Annovar":"Gene","Variant_Classification_Annovar":"Mutation"}) #Rename the columns we want to keep to the appropriate names
                    df['Location'] = df['Annovar_Info_protein'].str.extract(r'([^:]+$)') #The location that we care about is stored after the last colon
                    df = df[['Patient_ID', 'Gene', 'Mutation', 'Location']]

                elif self._version == "2.0":
                    df = df[['Tumor_Sample_Barcode','Hugo_Symbol','Variant_Classification','HGVSp_Short']]
                    df = df.rename(columns={
                        "Tumor_Sample_Barcode":"Patient_ID",
                        "Hugo_Symbol":"Gene",
                        "Variant_Classification":"Mutation",
                        "HGVSp_Short":"Location"}) #Rename the columns we want to keep to the appropriate names

                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                df = df.sort_index()
                df.columns.name=None
                self._data["somatic_mutation"] = df

            elif file_name == "clinic.tsi.gz" or file_name == "Meta_table.tsv.gz":
                df = pd.read_csv(file_path, sep="\t")

                if self._version == "2.0":
                    df = df.set_index('case_id')
                elif self._version == "0.1":
                    df = df.set_index('CASE_ID')

                df.columns.name=None
                df.index.name="Patient_ID"

                # Split the clinical data in to clinical data and derived molecular data

                if self._version == "0.1":
                    derived_molecular_cols = ['P53GENE_ANALYSIS', 'EGFR_AMP_STATUS']

                elif self._version == "2.0":
                    derived_molecular_cols = ['NAT_pathology_review', 'tumor_pathology_review',
                       'ESTIMATE_stromal_score', 'ESTIMATE_immune_score', 'stemness_score',
                       'mutation_count', 'TP53_mutation', 'CDKN2A_mutation', 'FAT1_mutation',
                       'NOTCH1_mutation', 'CSMD3_mutation', 'DNAH5_mutation', 'KMT2D_mutation',
                       'transcriptomic_subtype', 'chr_instability_idx', 'tumor_proportion',
                       'normal_epithelial_proportion', 'immune_proportion',
                       'muscle_proportion', 'fibroblast_proportion', 'EGFR_pathway',
                       'Hypoxia_pathway', 'JAK.STAT_pathway', 'MAPK_pathway', 'NFkB_pathway',
                       'PI3K_pathway', 'TGFb_pathway', 'TNFa_pathway', 'Trail_pathway',
                       'VEGF_pathway', 'p53_pathway']

                derived_molecular_df = df[derived_molecular_cols]
                derived_molecular_df = derived_molecular_df.sort_index(axis='columns')
                derived_molecular_df = derived_molecular_df.sort_index()

                df = df.drop(columns=derived_molecular_cols)
                df = df.sort_index()
                df = df.sort_index(axis='columns')

                self._data["clinical"] = df
                self._data["derived_molecular"] = derived_molecular_df

            elif file_name in ["Proteomics_DIA_Gene_level_Normal.cct.gz", "Proteomics_DIA_Gene_level_Tumor.cct.gz", "Proteomics_TMT_gene_level_combined_all.cct.gz"]:
                df = pd.read_csv(file_path, sep="\t")

                if self._version == "2.0":
                    df = df.set_index('Index')

                df = df.transpose()
                df.columns.name=None
                df.index.name = "Patient_ID"

                if self._version == "2.0":
                    df.index = df.index.str.replace(r'-T$', '', 1, regex=True)
                    df.index = df.index.str.replace(r'-N$', '.N', 1, regex=True)
                    df.index = df.index.str.replace(r'-C$', '.C', 1, regex=True) #-C is cored NAT samples

                # Once the files are formatted correctly load them into self._data
                if file_name == "Proteomics_DIA_Gene_level_Normal.cct.gz":
                    self._data["proteomics_normal"] = df

                elif file_name == "Proteomics_DIA_Gene_level_Tumor.cct.gz":
                    self._data["proteomics_tumor"] = df

                elif file_name == "Proteomics_TMT_gene_level_combined_all.cct.gz":
                    self._data["proteomics"] = df

            elif file_name == "Phosphoproteomics_TMT_site_level_combined_all.cct.gz" and self._version == "2.0":
                df = pd.read_csv(file_path, sep='\t')

                df = df.rename(columns={"Gene": "Name"})

                # Drop unlocalized sites
                unlocalized_sites = (df["Index"].str.rsplit("_", n=1, expand=True)[1] == '0')
                df = df[~unlocalized_sites]

                # Parse a few columns out of the "Index" column that we'll need for our multiindex
                split_ids = df["Index"].str.split('_', expand=True)
                df = df.drop(columns="Index")
                sites = split_ids.iloc[:, -1]
                database_ids = split_ids[0].str.cat(split_ids[1], sep='_')
                df = df.assign(**{"Site": sites, "Database_ID": database_ids})

                # Some rows have at least one localized phosphorylation site, but also have other phosphorylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~split_ids[4].eq(split_ids[5]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 4 of the split "Index" column is number of phosphorylations detected, and column 5 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"]) # This will create a multiindex from these columns, in this order.
                df = df.sort_index()
                df = df.transpose()
                df.index = df.index.str.replace(r'-T$', '', 1, regex=True)
                df.index = df.index.str.replace(r'-N$', '.N', 1, regex=True)
                df.index = df.index.str.replace(r'-C$', '.C', 1, regex=True) #-C is cored NAT samples
                df = df.sort_index()
                self._data["phosphoproteomics"] = df

            elif file_name == 'HN_followUp_9_24.xlsx' and self._version == "2.0":
                df = pd.read_excel(file_path)

                # Rename, set, and sort by index
                df = df.rename(columns={"CASE_ID": "Patient_ID"})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        if self._version == "0.1":
            # Combine the two proteomics dataframes
            df_normal = self._data.get("proteomics_normal")
            df_tumor = self._data.get("proteomics_tumor")

            df_normal.index = df_normal.index + ".N" #concatenate a ".N" onto the end of the normal data so we can identify it as normal after it's appended to tumor
            prot_combined = df_tumor.append(df_normal) #append the normal data onto the end of the tumor data
            prot_combined = prot_combined.sort_index(axis='columns') # Put all the columns in alphabetical order
            prot_combined = prot_combined.sort_index()
            self._data["proteomics"] = prot_combined
            del self._data["proteomics_normal"]
            del self._data["proteomics_tumor"]

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        master_clinical = self._data['clinical'].reindex(master_index)

        # Add a column called Sample_Tumor_Normal to the clinical dataframe indicating whether each sample is a tumor or normal sample. Samples with a Patient_ID ending in ".N" are normal.
        clinical_status_col = generate_sample_status_col(master_clinical, normal_test=lambda sample: sample[-2:] in ['.N', '.C'])
        master_clinical.insert(0, "Sample_Tumor_Normal", clinical_status_col)

        # Add a column called "Cored_Sample", with True for the six cored normal samples, and False for all others. The cored normal samples have a Patient_ID ending with a ".C".
        cored_sample_col = master_clinical.index.str.endswith(".C")
        master_clinical.insert(1, "Cored_Sample", cored_sample_col)

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = master_clinical

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='head and neck squamous cell carcinoma', pmid=33417831)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.file_tools import *
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError


class JoiningDataset(Dataset):

    def __init__(self, input_data):
        """Load all of the dataframes as values in the self._data dict variable, with names as keys, and format them properly.
        
        Parameters:
        input_data (dict of dict): Data dictionaries to combine for joining
        """

        # Call the parent class __init__ function
        super().__init__(cancer_type='joining', version=None, valid_versions=None, data_files=None, no_internet=None, skip_init=True)

        # Put all the input data tables into self._data
        # Each table's name needs to be its source followed by the data type, e.g. the proteomics table from bcm would be "bcm_transcriptomics"
               
        for source in input_data.keys():
            source_data = input_data[source]
            for data_type in source_data.keys():   
                self._data[source + "_" + data_type] = input_data[source][data_type]
        
        # create list of valid datasets ex bcm_transcriptomics, bcm_proteomics ect and put into self._valid_omics_dfs
        valid = []
        for source in input_data.keys():
            for df in self._valid_omics_dfs:
                valid.append(f"{source}_{df}")
                
        self._valid_omics_dfs = valid
        
        # create list of valid meta data sets ex: bcm_derived_molecular and put into self._valid_metadata_dfs
        meta_valid = []
        for source in input_data.keys():
            for df in self._valid_metadata_dfs:
                meta_valid.append(f"{source}_{df}")
                
        self._valid_metadata_dfs = meta_valid
        
  # for the join function to work the clincal data needs to be named clincal not mssm_clinical  
        self._data["clinical"] = self._data["mssm_clinical"]
        self._data["somatic_mutation"] = self._data["harmonized_somatic_mutation"]
        
        #create unionized indices for clinical
        
        master_index = unionize_indices(self._data).dropna()
        new_clinical = self._data["clinical"]
        new_clinical = new_clinical.reindex(master_index)
       
        sample_status_col = generate_sample_status_col(new_clinical, normal_test=lambda sample: sample.endswith('.N') if type(sample) is str else sample[0].endswith(".N"))
        #new_clinical = new_clinical.drop(columns=['Sample_Tumor_Normal'])
        
        new_clinical.insert(0, "Sample_Tumor_Normal", sample_status_col)
        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = new_clinical

        


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError

class Gbm(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the gbm dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0", "2.0", "2.1", "3.0"]

        data_files = {
            "1.0": [
                "clinical_data_core.v1.0.20190802.tsv.gz",
                "mirnaseq_mirna_mature_tpm.v1.0.20190802.tsv.gz",
                "phosphoproteome_pnnl_d6.v1.0.20190802.tsv.gz",
                "proteome_pnnl_per_gene_d4.v1.0.20190802.tsv.gz",
                "proteome_tmt_design.v1.0.20190802.tsv.gz",
                "rnaseq_gdc_fpkm_uq.v1.0.20190802.tsv.gz",
                "tindaisy_all_cases_filtered.v1.0.20190802.maf.gz",
                "wgs_somatic_cnv_per_gene.v1.0.20190802.tsv.gz"],
            "2.0": [
                "acetylome_pnnl_d6.v2.0.20190905.tsv.gz",
                "clinical_data_core.v2.0.20190905.tsv.gz",
                "metabolome_pnnl.v2.0.20190905.tsv.gz",
                "metabolome_sample_info.v2.0.20190905.tsv.gz",
                "mirnaseq_mirna_mature_tpm.v2.0.20190905.tsv.gz",
                "negative_lipidome_pnnl.v2.0.20190905.tsv.gz",
                "phosphoproteome_pnnl_d6.v2.0.20190905.tsv.gz",
                "positive_lipidome_pnnl.v2.0.20190905.tsv.gz",
                "proteome_pnnl_per_gene_d4.v2.0.20190905.tsv.gz",
                "proteome_tmt_design.v2.0.20190905.tsv.gz",
                "rnaseq_bcm_circular_rna_expression_rsem_uq.v2.0.20190905.tsv.gz",
                "rnaseq_gene_fusion.v2.0.20190905.tsv.gz",
                "rnaseq_washu_fpkm_uq.v2.0.20190905.tsv.gz",
                "tindaisy_all_cases_filtered.v2.0.20190905.maf.gz",
                "wgs_somatic_cnv_per_gene.v2.0.20190905.tsv.gz"],
            "2.1": [
                "acetylome_mssm_per_gene_clean.v2.1.20190927.tsv.gz",
                "clinical_data_core.v2.1.20190927.tsv.gz",
                "metabolome_pnnl.v2.1.20190927.tsv.gz",
                "metabolome_sample_info.v2.1.20190927.tsv.gz",
                "mirnaseq_mirna_mature_tpm.v2.1.20190927.tsv.gz",
                "negative_lipidome_pnnl.v2.1.20190927.tsv.gz",
                "phosphoproteome_mssm_per_gene_clean.v2.1.20190927.tsv.gz",
                "positive_lipidome_pnnl.v2.1.20190927.tsv.gz",
                "proteome_mssm_per_gene_clean.v2.1.20190927.tsv.gz",
                "proteome_tmt_design.v2.1.20190927.tsv.gz",
                "rnaseq_bcm_circular_rna_expression_rsem_uq.v2.1.20190927.tsv.gz",
                "rnaseq_gene_fusion.v2.1.20190927.tsv.gz",
                "rnaseq_washu_fpkm_uq.v2.1.20190927.tsv.gz",
                "tindaisy_all_cases_filtered.v2.1.20190927.maf.gz",
                "wgs_somatic_cnv_per_gene.v2.1.20190927.tsv.gz"],
            "3.0": [
                "acetylome_mssm_per_gene_clean.v3.0.20191121.tsv.gz",
                "clinical_data_core.v3.0.20191121.tsv.gz",
                "gbm_all_subtype_collections.2020-01-13.tsv.gz",
                "metabolome_pnnl.v3.0.20191121.tsv.gz",
                "metabolome_sample_info.v3.0.20191121.tsv.gz",
                "mirnaseq_mirna_mature_tpm.v3.0.20191121.tsv.gz",
                "negative_lipidome_pnnl.v3.0.20191121.tsv.gz",
                "phosphoproteome_mssm_per_gene_clean.v3.0.20191121.tsv.gz",
                "positive_lipidome_pnnl.v3.0.20191121.tsv.gz",
                "proteome_mssm_per_gene_clean.v3.0.20191121.tsv.gz",
                "proteome_tmt_design.v3.0.20191121.tsv.gz",
                "rnaseq_bcm_circular_rna_expression_rsem_uq.v3.0.20191121.tsv.gz",
                "rnaseq_gene_fusion.v3.0.20191121.tsv.gz",
                "rnaseq_washu_fpkm_uq.v3.0.20191121.tsv.gz",
                "tindaisy_all_cases_filtered.v3.0.20191121.maf.gz",
                "wgs_somatic_cnv_per_gene.v3.0.20191121.tsv.gz"],
        }

        super().__init__(cancer_type="gbm", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Our dataframe name will be the first section of file name, so we don't include the version

            if df_name in ("acetylome_pnnl_d6", "acetylome_mssm_per_gene_clean"):
                df = pd.read_csv(file_path, sep='\t')
                split_genes = df["site"].str.rsplit("-", n=1,expand=True)  # Split the genes from the sites, splitting from the right since some genes have hyphens in their names, but the genes and sites are also separated by hyphens
                df = df.drop(columns="site")
                df = df.assign(Site=split_genes[1])
                df["Site"] = df["Site"].str.replace(r"k", r"", regex=True)  # Get rid of all lowercase k delimeters in the sites

                # Create the multiindex
                df = df.rename(columns={
                        "gene": "Name",
                        "peptide": "Peptide",
                        "refseq_id": "Database_ID",
                    })
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])  # Turn these columns into a multiindex
                df = df.sort_index()

                df = df.transpose()
                self._data["acetylproteomics"] = df

            elif df_name == "clinical_data_core":
                df = pd.read_csv(file_path, sep='\t', index_col=0).\
                    assign(Stage="IV") # By definition they're all stage IV, since it's glioblastoma
                self._data["clinical"] = df

            elif file_name == "gbm_all_subtype_collections.2020-01-13.tsv.gz":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.drop(columns="sample_type")
                self._data["derived_molecular"] = df

            elif df_name == "metabolome_pnnl":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.transpose()
                self._data["metabolomics"] = df

            elif df_name == "metabolome_sample_info":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df.index.name = "Patient_ID"
                self._data["sample_info"] = df

            elif df_name == "mirnaseq_mirna_mature_tpm":
                df = pd.read_csv(file_path, sep='\t')
                df = df.rename(columns={"name": "Name", "unique_id": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"]) # We use a multiindex with database IDs, not just names, to avoid duplicate column headers
                df = df.drop(columns=["chromosome", "start", "end", "strand", "mirna_type", "mirbase_id", "precursor_id"])
                df = df.sort_index()
                df = df.transpose()
                self._data["miRNA"] = df

            elif df_name == "negative_lipidome_pnnl":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.transpose()
                df = df.add_suffix("_negative")
                self._data["lipidomics_negative"] = df

            elif df_name in ("phosphoproteome_pnnl_d6", "phosphoproteome_mssm_per_gene_clean"):
                df = pd.read_csv(file_path, sep='\t')

                # Create our multiindex
                split_genes = df["site"].str.rsplit("-", n=1, expand=True) # Split the genes from the sites, splitting from the right since some genes have hyphens in their names, but the genes and sites are also separated by hyphens
                df = df.drop(columns="site")
                df = df.assign(Site=split_genes[1])
                df["Site"] = df["Site"].str.replace(r"[sty]", r"", regex=True) # Get rid of all lowercase s, t, and y delimeters in the sites

                if self._version == "1.0":
                    df = df.rename(columns={"gene": "Name", "peptide": "Peptide"})
                    df = df.set_index(["Name", "Site", "Peptide"]) # Turn these columns into a multiindex

                elif self._version in ("2.0", "2.1", "3.0"):
                    df = df.rename(columns={
                            "gene": "Name",
                            "peptide": "Peptide",
                            "refseq_id": "Database_ID",
                        })
                    df = df.set_index(["Name", "Site", "Peptide", "Database_ID"]) # Turn these columns into a multiindex

                df = df.sort_index()
                df = df.transpose()
                self._data["phosphoproteomics"] = df

            elif df_name == "positive_lipidome_pnnl":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.transpose()
                df = df.add_suffix("_positive")
                self._data["lipidomics_positive"] = df

            elif df_name in ("proteome_pnnl_per_gene_d4", "proteome_mssm_per_gene_clean"):
                df = pd.read_csv(file_path, sep='\t', index_col=0)

                if self._version in ("2.0", "2.1", "3.0"):
                    df = df.drop(columns="refseq_id") # We don't need this database ID, because the gene name index is already unique

                df = df.sort_index()
                df = df.transpose()
                self._data["proteomics"] = df

            elif df_name == "proteome_tmt_design":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df.index.name = "Patient_ID"
                self._data["experimental_design"] = df

            elif df_name == "rnaseq_bcm_circular_rna_expression_rsem_uq":
                df = pd.read_csv(file_path, sep='\t')
                df["circRNA_id"] = df["circRNA_id"].str.split('_', n=1, expand=True)[1] # Drop the "circ_" prefix on all the keys
                df = df.set_index("circRNA_id")
                df = df.drop(columns=["gene_id", "gene_name", "gene_type", "alias"])
                df = df.transpose()
                self._data["circular_RNA"] = df

            elif df_name == "rnaseq_gene_fusion":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                self._data["gene_fusion"] = df

            elif df_name in ("rnaseq_gdc_fpkm_uq", "rnaseq_washu_fpkm_uq"):
                df = pd.read_csv(file_path, sep='\t')
                df = df.rename(columns={"gene_name": "Name", "gene_id": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"]) # We use a multiindex with Ensembl IDs, not just gene names, to avoid duplicate column headers
                df = df.drop(columns=["gene_type", "gene_status", "havana_gene", "full_length", "exon_length", "exon_num"])
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data["transcriptomics"] = df

            elif df_name == "tindaisy_all_cases_filtered":
                df = pd.read_csv(file_path, sep='\t')
                split_barcode = df["Tumor_Sample_Barcode"].str.split("_", n = 1, expand = True) # The first part of the barcode is the patient id, which we need want to make the index
                df["Tumor_Sample_Barcode"] = split_barcode[0]
                df = df[["Tumor_Sample_Barcode","Hugo_Symbol","Variant_Classification","HGVSp_Short"]]
                df = df.rename({"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol":"Gene","Variant_Classification":"Mutation","HGVSp_Short":"Location"}, axis='columns')
                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                self._data["somatic_mutation"] = df

            elif df_name == "wgs_somatic_cnv_per_gene":
                df = pd.read_csv(file_path, sep='\t', index_col=0)
                df = df.drop(columns=["gene_id", "gene_id_version", "original_symbol"])
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data["CNV"] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        if self._version in ("2.0", "2.1", "3.0"):
            # Combine positive and negative lipidomics tables
            lipidomics_positive = self._data["lipidomics_positive"]
            lipidomics_negative = self._data["lipidomics_negative"]

            lipidomics = lipidomics_positive.join(lipidomics_negative, how="outer")
            self._data["lipidomics"] = lipidomics

            del self._data["lipidomics_positive"]
            del self._data["lipidomics_negative"]

            # Add useful columns from sample_info table to experimental_design
            sample_info = self._data["sample_info"]
            useful_cols = sample_info[["mass_mg", "is_oct"]]
            useful_cols = useful_cols.add_prefix("sample_")

            experimental_design = self._data["experimental_design"]
            experimental_design = experimental_design.join(useful_cols, how="outer")

            self._data["experimental_design"] = experimental_design
            del self._data["sample_info"]

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data)

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)

        # Construct the sample status column
        sample_status_col = np.where(clinical.index.str.startswith("PT"), "Normal", "Tumor")
        clinical.insert(0, "Sample_Tumor_Normal", sample_status_col)

        # For versions 1.0, 2.0, and 2.1, the gender is mis-entered for two samples in the clinical dataframe. Both C3N-01196 and C3N-01856 are entered as Female, but are actually Male. Let's fix that, if we're loading one of those versions.
        if self._version in ("1.0", "2.0", "2.1"):
            clinical.loc[clinical.index.isin(["C3N-01196", "C3N-01856"]), "gender"] = "Male"

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = clinical

        # Append a ".N" to the Patient_IDs of the normal samples, to match the other datasets
        self._data = reformat_normal_patient_ids(self._data)

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

        # Print data embargo warning, if the date hasn't passed yet.
        today = datetime.date.today()
        embargo_date = datetime.date(year=2021, month=3, day=1)
        if today < embargo_date:
            warnings.warn("The GBM dataset is under publication embargo until March 01, 2021. CPTAC is a community resource project and data are made available rapidly after generation for community research use. The embargo allows exploring and utilizing the data, but analysis may not be published until after the embargo date. Please see https://proteomics.cancer.gov/data-portal/about/data-use-agreement or enter cptac.embargo() to open the webpage for more details.", PublicationEmbargoWarning, stacklevel=2)

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='glioblastoma', pmid=33577785)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
from .dataset import Dataset
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError, PublicationEmbargoWarning

class Lscc(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the lscc dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["1.0", "3.2", "3.2.1", "3.3"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "1.0": [
                "lscc-v1.0-cnv-gene-level-log2.gct.gz",
                "lscc-v1.0-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz",
                "lscc-v1.0-cptac3-lscc-wxs-somatic-variant-sw-v1.5-lscc.y2-20191211.maf.gz",
                "lscc-v1.0-mirna-mature-tpm-log2.gct.gz",
                "lscc-v1.0-phosphoproteome-ratio-norm-NArm.gct.gz",
                "lscc-v1.0-proteome-ratio-norm-NArm.gct.gz",
                "lscc-v1.0-rnaseq-uq-fpkm-log2-NArm.gct.gz",
                "lscc-v1.0-sample-annotation.csv.gz"],
            "2.0": [
                "lscc-v2.0-phosphoproteome-ratio-norm-NArm.gct.gz",
                "lscc-v2.0-proteome-ratio-norm-NArm.gct.gz",
                "lscc-v2.0-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz",
                "lscc-v2.0-mirna-mature-tpm-log2.gct.gz",
                "lscc-v2.0-rnaseq-uq-fpkm-log2-NArm.gct.gz",
                "lscc-v2.0-acetylome-ratio-norm-NArm.gct.gz",
                "lscc-v2.0-sample-annotation.csv.gz",
                "lscc-v2.0-gene-level-cnv-gistic2-all_data_by_genes.gct.gz",
                "lscc-v2.0-cptac3-lscc-wxs-somatic-v2.1-lscc.20191228-20200107-maf-like.txt.gz"],
            "3.2": [
                "lscc-v3.2-phosphoproteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-proteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz",
                "lscc-v3.2-mirna-mature-tpm-log2.gct.gz",
                "lscc-v3.2-rnaseq-uq-fpkm-log2-NArm.gct.gz",
                "lscc-v3.2-acetylome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-sample-annotation.csv.gz",
                "lscc-v3.2-gene-level-cnv-gistic2-all_data_by_genes.gct.gz",
                "lscc-v3.2-mutsig-2cv-umich-v2-lscc-poncptac3-lscc-v3beta.final-analysis-set.maf.gz",
                "lscc-v3.2-ubiquitylome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-circular-rna-rsem-uq-log2.gct.gz"],
            "3.2.1": [
                "lscc-v3.2-acetylome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-circular-rna-rsem-uq-log2.gct.gz",
                "lscc-v3.2-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz",
                "lscc-v3.2-gene-level-cnv-gistic2-log-ratio-all_data_by_genes.gct.gz",
                "lscc-v3.2-mirna-mature-tpm-log2.gct.gz",
                "lscc-v3.2-mutsig-2cv-umich-v2-lscc-poncptac3-lscc-v3beta.final-analysis-set.maf.gz",
                "lscc-v3.2-phosphoproteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-proteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.2-rnaseq-uq-fpkm-log2-NArm.gct.gz",
                "lscc-v3.2-sample-annotation.csv.gz",
                "lscc-v3.2-ubiquitylome-ratio-norm-NArm.gct.gz"],
            "3.3": [
                "lscc-v3.2-mutsig-2cv-umich-v2-lscc-poncptac3-lscc-v3beta.final-analysis-set.maf.gz",
                "lscc-v3.2-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz",
                "lscc-v3.3-public-acetylome-ratio-norm-NArm.gct.gz",
                "lscc-v3.3-public-circular-rna-rsem-uq-log2.gct.gz",
                "lscc-v3.3-public-gene-level-cnv-gistic2-log-ratio-all_data_by_genes.gct.gz",
                "lscc-v3.3-public-methylation-promoter-5utr-gene-level.gct.gz",
                "lscc-v3.3-public-mirna-mature-tpm-log2.gct.gz",
                "lscc-v3.3-public-phosphoproteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.3-public-proteome-ratio-norm-NArm.gct.gz",
                "lscc-v3.3-public-rnaseq-uq-fpkm-log2-NArm.gct.gz",
                "lscc-v3.3-public-sample-annotation.csv.gz",
                "lscc-v3.3-public-ubiquitylome-batch-corrected-ratio-norm-NArm.gct.gz"],
        }

        super().__init__(cancer_type="lscc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file

            if file_name == "lscc-v1.0-cnv-gene-level-log2.gct.gz":
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                df = df.set_index("id")
                cols_to_drop = ["chrom","geneSymbol","chr_end","chr_start"]
                df = df.drop(columns = cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                self._data["CNV"] = df

            elif file_name in ["lscc-v1.0-phosphoproteome-ratio-norm-NArm.gct.gz", "lscc-v2.0-phosphoproteome-ratio-norm-NArm.gct.gz", "lscc-v3.2-phosphoproteome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-phosphoproteome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Drop rows of metadata
                df = df[gene_filter]

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                "geneSymbol": "Name",
                "variableSites": "Site",
                "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                })

                # Some rows have at least one localized phosphorylation site, but also have other phosphorylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~df['Best_numActualVMSites_sty'].eq(df['Best_numLocalizedVMsites_sty']) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split "id" column is number of phosphorylations detected, and column 4 is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])

                cols_to_drop = ['id','id.description', 'numColumnsVMsiteObserved', 'bestScore', 'bestDeltaForwardReverseScore',
                'Best_scoreVML', 'Best_numActualVMSites_sty', 'Best_numLocalizedVMsites_sty', 'sequenceVML',
                'accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA', 'protein_mw', 'species',
                'speciesMulti', 'orfCategory', 'accession_number', 'protein_group_num', 'entry_name', 'GeneSymbol']

                if file_name in ["lscc-v2.0-phosphoproteome-ratio-norm-NArm.gct.gz","lscc-v3.2-phosphoproteome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-phosphoproteome-ratio-norm-NArm.gct.gz"]:
                    cols_to_drop.extend(['VMsiteFlanks','Best_numAmbiguousVMsites_sty', 'StartAA'])
                df = df.drop(columns=cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                self._data["phosphoproteomics"] = df

            elif file_name in ["lscc-v1.0-proteome-ratio-norm-NArm.gct.gz", "lscc-v2.0-proteome-ratio-norm-NArm.gct.gz","lscc-v3.2-proteome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-proteome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, skiprows=2, sep='\t', dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]

                df = df.rename(columns={"GeneSymbol": "Name", 'accession_numbers': "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                cols_to_drop = ['id', 'id.description', 'geneSymbol', 'numColumnsProteinObserved',
                'numSpectraProteinObserved', 'protein_mw', 'percentCoverage', 'numPepsUnique',
                'scoreUnique', 'species', 'orfCategory', 'accession_number',
                'subgroupNum', 'entry_name']
                if file_name in ["lscc-v2.0-proteome-ratio-norm-NArm.gct.gz", "lscc-v3.2-proteome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-proteome-ratio-norm-NArm.gct.gz"]:
                    cols_to_drop.extend(['numPepsUniqueSubgroupSpecificCI', 'scoreUniqueSubgroupSpecificCI'])
                df = df.drop(columns=cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                df.columns.name=None
                self._data["proteomics"] = df


            elif file_name in ["lscc-v1.0-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz", "lscc-v2.0-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz","lscc-v3.2-cptac3-lscc-rna-seq-fusion-v2.2-y2.all-20190807.txt.gz"]:
                df = pd.read_csv(file_path, sep="\t", dtype=object)
                df = df.rename(columns={"Sample.ID": "Patient_ID"})
                df = df.set_index("Patient_ID")

                self._data['gene_fusion'] = df

            elif file_name == "lscc-v1.0-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, sep=",", dtype=object)
                filter = df['QC.status'] == "QC.pass" #There are some samples that are internal references. IRs are used for scaling purposes, and don't belong to a single patient, so we want to drop them.
                df = df[filter]
                df = df.drop(columns="Participant") #Get rid of the "Participant" column becuase the same information is stored in Sample.ID  which is formatted the way we want.
                df = df.set_index("Sample.ID")
                df = df.drop(columns="Sample.IDs")
                df.index.name="Patient_ID"
                df = df.rename(columns={"Type":"Sample_Tumor_Normal"})
                df["Sample_Tumor_Normal"] = df["Sample_Tumor_Normal"].replace("NAT","Normal")

                #Split the metadata into multiple dataframes
                #Make experiemntal_set up dataframe
                experimental_design_cols = ['Experiment', 'Channel', 'QC.status'] #These are the columns for the experimental_design dataframe
                experimental_design_df = df[experimental_design_cols]
                df = df.drop(columns=experimental_design_cols)

                #Make a derived_molecular dataframe
                derived_molecular_cols = ['TP53.mutation', 'CDKN2A.mutation', 'PTEN.mutation', 'PIK3CA.mutation',
                 'KEAP1.mutation', 'HLA.A.mutation', 'NFE2L2.mutation', 'NOTCH1.mutation', 'RB1.mutation',
                 'HRAS.mutation', 'FBXW7.mutation', 'SMARCA4.mutation', 'NF1.mutation', 'SMAD4.mutation',
                 'EGFR.mutation', 'APC.mutation', 'BRAF.mutation', 'TNFAIP3.mutation', 'CREBBP.mutation',
                 'TP53.mutation.status', 'CDKN2A.mutation.status', 'PTEN.mutation.status', 'PIK3CA.mutation.status',
                 'KEAP1.mutation.status', 'HLA.A.mutation.status', 'NFE2L2.mutation.status', 'NOTCH1.mutation.status',
                 'RB1.mutation.status', 'HRAS.mutation.status', 'FBXW7.mutation.status', 'SMARCA4.mutation.status',
                 'NF1.mutation.status', 'SMAD4.mutation.status', 'EGFR.mutation.status', 'APC.mutation.status',
                 'BRAF.mutation.status', 'TNFAIP3.mutation.status', 'CREBBP.mutation.status']
                derived_molecular_df = df[derived_molecular_cols]
                df = df.drop(columns = derived_molecular_cols)
                self._data["clinical"]= df
                self._data['experimental_design'] = experimental_design_df
                self._data['derived_molecular'] = derived_molecular_df

            elif file_name == "lscc-v2.0-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, sep=",", dtype=object)
                filter = df['QC.status'] == "QC.pass" #There are some samples that are internal references. IRs are used for scaling purposes, and don't belong to a single patient, so we want to drop them.
                df = df[filter]
                df = df.drop(columns="Participant") #Get rid of the "Participant" column becuase the same information is stored in Sample.ID  which is formatted the way we want.
                df = df.set_index("Sample.ID")
                df = df.drop(columns="Sample.IDs")
                df.index.name="Patient_ID"
                df = df.rename(columns={"Type":"Sample_Tumor_Normal"})
                df["Sample_Tumor_Normal"] = df["Sample_Tumor_Normal"].replace("NAT","Normal")

                #Split the metadata into multiple dataframes
                #Make experiemntal_set up dataframe
                experimental_design_cols = ['Experiment', 'Channel', 'QC.status','Aliquot'] #These are the columns for the experimental_design dataframe
                experimental_design_df = df[experimental_design_cols]
                df = df.drop(columns=experimental_design_cols)

                #Make a derived_molecular dataframe
                derived_molecular_cols = ['TP53.mutation', 'CDKN2A.mutation', 'PTEN.mutation', 'KMT2D.mutation',
                       'NFE2L2.mutation', 'ARID1A.mutation', 'NOTCH1.mutation', 'NF1.mutation',
                       'CUL3.mutation', 'KEAP1.mutation', 'KRAS.mutation', 'KDM6A.mutation',
                       'RANBP2.mutation', 'TP53.mutation.status', 'CDKN2A.mutation.status',
                       'PTEN.mutation.status', 'KMT2D.mutation.status',
                       'NFE2L2.mutation.status', 'ARID1A.mutation.status',
                       'NOTCH1.mutation.status', 'NF1.mutation.status', 'CUL3.mutation.status',
                       'KEAP1.mutation.status', 'KRAS.mutation.status',
                       'KDM6A.mutation.status', 'RANBP2.mutation.status', 'CIN.wxs',
                       'Subtype.TCGA.rna', 'NMF.cluster', 'NMF.cluster.membership.score',
                       'Smoking.Signature.Fraction.wxs', 'Smoking.Signature.Count.wxs',
                       'Total.Mutation.Count.wxs', 'Mutation.Count.ExcludingINDELs.wxs',
                       'DNP.Count.wxs', 'DNP.Count.GG.to.TT.or.CC.to.AA.wxs',
                       'Smoking.score.wxs', 'Smoking.Score.Category.wxs',
                       'Mutation.Count.Excluding.Silent.wxs',
                       'Total.Mutation.Count.per.Mb.wxs', 'ESTIMATE.StromalScore.rna',
                       'ESTIMATE.ImmuneScore.rna', 'ESTIMATE.TumorPurity.rna',
                       'TSNet.Purity.rna', 'Immune.Cluster.rna', 'xCell.ImmuneScore.rna',
                       'xCell.StromaScore.rna', 'xCell.MicroenvironmentScore.rna',
                       'CIBERSORT.AbsoluteScore.rna']
                derived_molecular_df = df[derived_molecular_cols]
                df = df.drop(columns = derived_molecular_cols)
                self._data["clinical"]= df
                self._data['experimental_design'] = experimental_design_df
                self._data['derived_molecular'] = derived_molecular_df

            elif file_name == "lscc-v3.2-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, sep=",", dtype=object)
                filter = df['QC.status'] == "QC.pass" #There are some samples that are internal references. IRs are used for scaling purposes, and don't belong to a single patient, so we want to drop them.
                df = df[filter]
                df = df.drop(columns="Participant") #Get rid of the "Participant" column becuase the same information is stored in Sample.ID  which is formatted the way we want.
                df = df.set_index("Sample.ID")
                df = df.drop(columns="Sample.IDs")
                df.index.name="Patient_ID"
                df = df.rename(columns={"Type":"Sample_Tumor_Normal"})
                df["Sample_Tumor_Normal"] = df["Sample_Tumor_Normal"].replace("NAT","Normal")
                #Split the metadata into multiple dataframes
                #Make experiemntal_set up dataframe
                experimental_design_cols = ['Experiment', 'Channel', 'QC.status','Aliquot.tmt'] #These are the columns for the experimental_design dataframe
                experimental_design_df = df[experimental_design_cols]
                df = df.drop(columns=experimental_design_cols)
                # #Make a derived_molecular dataframe
                derived_molecular_cols = ['CIMP.status.meth','TP53.mutation','PTEN.mutation',
                 'CDKN2A.mutation','KMT2D.mutation','NFE2L2.mutation','ARID1A.mutation',
                 'CUL3.mutation','BRCA2.mutation','KEAP1.mutation','SUZ12.mutation','NF1.mutation',
                 'PIK3CA.mutation','NOTCH1.mutation','TP53.mutation.status','PTEN.mutation.status','CDKN2A.mutation.status',
                 'KMT2D.mutation.status','NFE2L2.mutation.status','ARID1A.mutation.status','CUL3.mutation.status',
                 'BRCA2.mutation.status','KEAP1.mutation.status','SUZ12.mutation.status','NF1.mutation.status',
                 'PIK3CA.mutation.status','NOTCH1.mutation.status','CUL3.NFE2L2.KEAP1.mutation.status','KAT6A.scna.wxs',
                 'SOX2.scna.wxs','TP63.scna.wxs','FGFR1.scna.wxs','CDKN2A.scna.wxs','CDKN2A.pathway.alteration',
                 'CDKN2A.pathway.alteration.status','PIK3CA.pathway.alteration','PIK3CA.pathway.alteration.status',
                 'FGFR3.TACC3.fusion.rna','Subtype.TCGA.rna','NMF.cluster','NMF.cluster.core',
                 'NMF.cluster.membership.score','CIN.wxs','Smoking.Signature.Fraction.wxs','Smoking.Signature.Count.wxs',
                 'Total.Mutation.Count.wxs','Mutation.Count.ExcludingINDELs.wxs','DNP.Count.wxs','DNP.Count.GG.to.TT.or.CC.to.AA.wxs',
                 'Smoking.score.wxs','Smoking.Score.Category.wxs','Mutation.Count.Excluding.Silent.wxs','Total.Mutation.Count.per.Mb.wxs',
                 'Immune.Subtype.Thorsson2018.rna','ESTIMATE.StromalScore.rna','ESTIMATE.ImmuneScore.rna','ESTIMATE.TumorPurity.rna',
                 'TSNet.Purity.rna','Immune.Cluster.rna','xCell.ImmuneScore.rna','xCell.StromaScore.rna','xCell.MicroenvironmentScore.rna',
                 'CIBERSORT.AbsoluteScore.rna']
                derived_molecular_df = df[derived_molecular_cols]
                df = df.drop(columns = derived_molecular_cols)
                self._data["clinical"]= df
                self._data['experimental_design'] = experimental_design_df
                self._data['derived_molecular'] = derived_molecular_df
                
            elif file_name == "lscc-v3.3-public-sample-annotation.csv.gz":
                df = pd.read_csv(file_path, sep=",")
                filter = df['QC.status'] == "QC.pass" #There are some samples that are internal references. IRs are used for scaling purposes, and don't belong to a single patient, so we want to drop them.
                df = df[filter]
                df = df.drop(columns="Participant") #Get rid of the "Participant" column becuase the same information is stored in Sample.ID  which is formatted the way we want.
                df = df.set_index("Sample.ID")
                df = df.drop(columns="Sample.IDs")
                df.index.name="Patient_ID"
                df = df.rename(columns={"Type":"Sample_Tumor_Normal"})
                df["Sample_Tumor_Normal"] = df["Sample_Tumor_Normal"].replace("NAT","Normal")
                #Split the metadata into multiple dataframes
                #Make experiemntal_set up dataframe
                experimental_design_cols = ['Experiment', 'Channel', 'QC.status','Aliquot.tmt'] #These are the columns for the experimental_design dataframe
                experimental_design_df = df[experimental_design_cols]
                df = df.drop(columns=experimental_design_cols)
                # #Make a derived_molecular dataframe
                derived_molecular_cols = ['CIMP.status.meth','TP53.mutation','PTEN.mutation',
                 'CDKN2A.mutation','KMT2D.mutation','NFE2L2.mutation','ARID1A.mutation',
                 'CUL3.mutation','BRCA2.mutation','KEAP1.mutation','SUZ12.mutation','NF1.mutation',
                 'PIK3CA.mutation','NOTCH1.mutation','TP53.mutation.status','PTEN.mutation.status','CDKN2A.mutation.status',
                 'KMT2D.mutation.status','NFE2L2.mutation.status','ARID1A.mutation.status','CUL3.mutation.status',
                 'BRCA2.mutation.status','KEAP1.mutation.status','SUZ12.mutation.status','NF1.mutation.status',
                 'PIK3CA.mutation.status','NOTCH1.mutation.status','CUL3.NFE2L2.KEAP1.mutation.status','KAT6A.scna.wxs',
                 'SOX2.scna.wxs','TP63.scna.wxs','FGFR1.scna.wxs','CDKN2A.scna.wxs','CDKN2A.pathway.alteration',
                 'CDKN2A.pathway.alteration.status','PIK3CA.pathway.alteration','PIK3CA.pathway.alteration.status',
                 'FGFR3.TACC3.fusion.rna','Subtype.TCGA.rna','NMF.cluster','NMF.cluster.core',
                 'NMF.cluster.membership.score','CIN.wxs','Smoking.Signature.Fraction.wxs','Smoking.Signature.Count.wxs',
                 'Total.Mutation.Count.wxs','Mutation.Count.ExcludingINDELs.wxs','DNP.Count.wxs','DNP.Count.GG.to.TT.or.CC.to.AA.wxs',
                 'Smoking.score.wxs','Smoking.Score.Category.wxs','Mutation.Count.Excluding.Silent.wxs','Total.Mutation.Count.per.Mb.wxs',
                 'Immune.Subtype.Thorsson2018.rna','ESTIMATE.StromalScore.rna','ESTIMATE.ImmuneScore.rna','ESTIMATE.TumorPurity.rna',
                 'TSNet.Purity.rna','Immune.Cluster.rna','xCell.ImmuneScore.rna','xCell.StromaScore.rna','xCell.MicroenvironmentScore.rna',
                 'CIBERSORT.AbsoluteScore.rna']
                derived_molecular_df = df[derived_molecular_cols]
                df = df.drop(columns = derived_molecular_cols)
                self._data["clinical"]= df
                self._data['experimental_design'] = experimental_design_df
                self._data['derived_molecular'] = derived_molecular_df

            elif file_name == "lscc-v1.0-cptac3-lscc-wxs-somatic-variant-sw-v1.5-lscc.y2-20191211.maf.gz":
                df = pd.read_csv(file_path, sep="\t", dtype=object)
                df = df[["Sample.ID", "Hugo_Symbol", "Variant_Classification", "HGVSp_Short"]] # We don't need any of the other columns
                df = df.rename(columns={"Sample.ID": "Patient_ID", 'Hugo_Symbol': "Gene", "Variant_Classification": "Mutation", "HGVSp_Short": "Location"})
                df = df.set_index("Patient_ID")
                df = df.sort_values(by=["Patient_ID","Gene"])
                self._data['somatic_mutation'] = df

            elif file_name == "lscc-v2.0-cptac3-lscc-wxs-somatic-v2.1-lscc.20191228-20200107-maf-like.txt.gz":
                df = pd.read_csv(file_path, sep="\t", dtype=object)
                df = df[["Sample.ID", "Hugo_Symbol", "Variant_Classification", "POS"]] # We don't need any of the other columns
                df = df.rename(columns={"Sample.ID": "Patient_ID", 'Hugo_Symbol': "Gene", "Variant_Classification": "Mutation", "POS": "Location"})
                df = df.set_index("Patient_ID")
                df = df.sort_values(by=["Patient_ID","Gene"])
                self._data['somatic_mutation'] = df

            elif file_name == "lscc-v3.2-mutsig-2cv-umich-v2-lscc-poncptac3-lscc-v3beta.final-analysis-set.maf.gz":
                df = pd.read_csv(file_path, sep="\t", dtype=object)
                df = df[["Sample.ID", "Hugo_Symbol", "Variant_Classification", "Protein_Change"]] # We don't need any of the other columns
                df = df.rename(columns={"Sample.ID": "Patient_ID", 'Hugo_Symbol': "Gene", "Variant_Classification": "Mutation", "Protein_Change": "Location"})
                df = df.set_index("Patient_ID")
                df = df.sort_values(by=["Patient_ID","Gene"])
                self._data['somatic_mutation'] = df

            elif file_name in ["lscc-v1.0-mirna-mature-tpm-log2.gct.gz","lscc-v2.0-mirna-mature-tpm-log2.gct.gz", "lscc-v3.2-mirna-mature-tpm-log2.gct.gz", "lscc-v3.3-public-mirna-mature-tpm-log2.gct.gz"]:
                df = pd.read_csv(file_path, skiprows=2, sep='\t', dtype=object)
                gene_filter = df['Name'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                df = df.rename(columns={"ID": "Database_ID"})
                df = df.set_index(["Name","Database_ID"])
                cols_to_drop = ["Derives_from","Quantified.in.Percent.Samples","id","Alias"]
                df = df.drop(columns = cols_to_drop)
                df= df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                df.columns.name=None
                self._data["miRNA"] = df

            elif file_name in ["lscc-v1.0-rnaseq-uq-fpkm-log2-NArm.gct.gz","lscc-v2.0-rnaseq-uq-fpkm-log2-NArm.gct.gz", "lscc-v3.2-rnaseq-uq-fpkm-log2-NArm.gct.gz", "lscc-v3.3-public-rnaseq-uq-fpkm-log2-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep="\t", dtype=object,skiprows=2)
                gene_filter = df['gene_id'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                df = df.set_index("id")
                cols_to_drop = ["ENSEMBL","geneSymbol","GENENAME","gene_id"]
                df = df.drop(columns = cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df.sort_index()
                df.index.name="Patient_ID"
                self._data["transcriptomics"] = df

            elif file_name in ["lscc-v2.0-acetylome-ratio-norm-NArm.gct.gz","lscc-v3.2-acetylome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-acetylome-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Drop rows of metadata
                df = df[gene_filter]

                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "geneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })


                # Some rows have at least one localized acetylation site, but also have other acetylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                localization = df["accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA"].str.split("_", expand=True)
                unlocalized_to_drop = localization.index[~localization[3].eq(localization[4]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split localization data column is number of acetylation sites detected, and column 4 is number of acetylation sites localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])

                cols_to_drop = [
                    "id",
                    "id.description",
                    "numColumnsVMsiteObserved",
                    "bestScore",
                    "bestDeltaForwardReverseScore",
                    "Best_scoreVML",
                    "sequenceVML",
                    "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA",
                    "protein_mw",
                    "species",
                    "speciesMulti",
                    "orfCategory",
                    "accession_number",
                    "protein_group_num",
                    "entry_name",
                    "GeneSymbol",
                    "VMsiteFlanks",
                    "Best_numActualVMSites_k",
                    "Best_numAmbiguousVMsites_k",
                    "Best_numLocalizedVMsites_k",
                    "StartAA"]

                df = df.drop(columns=cols_to_drop)

                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"

                self._data["acetylproteomics"] = df

            elif file_name in ["lscc-v2.0-gene-level-cnv-gistic2-all_data_by_genes.gct.gz","lscc-v3.2-gene-level-cnv-gistic2-all_data_by_genes.gct.gz", "lscc-v3.2-gene-level-cnv-gistic2-log-ratio-all_data_by_genes.gct.gz", "lscc-v3.3-public-gene-level-cnv-gistic2-log-ratio-all_data_by_genes.gct.gz"]:
                df = pd.read_csv(file_path, sep="\t", skiprows=2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                df = df.set_index("id")
                cols_to_drop = ["Cytoband","Gene.ID","geneSymbol"]
                if file_name in ["lscc-v3.2-gene-level-cnv-gistic2-all_data_by_genes.gct.gz", "lscc-v3.2-gene-level-cnv-gistic2-log-ratio-all_data_by_genes.gct.gz"]:
                    cols_to_drop.extend(['Samples.altered','Samples.amp','Samples.del'])
                df = df.drop(columns=cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                self._data["CNV"] = df

            elif file_name in ["lscc-v3.2-ubiquitylome-ratio-norm-NArm.gct.gz", "lscc-v3.3-public-ubiquitylome-batch-corrected-ratio-norm-NArm.gct.gz"]:
                df = pd.read_csv(file_path, skiprows=2, sep='\t', dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                # Prepare some columns we'll need later for the multiindex
                df["variableSites"] = df["variableSites"].str.replace(r"[a-z\s]", "", regex=True) # Get rid of all lowercase delimeters and whitespace in the sites
                df = df.rename(columns={
                    "geneSymbol": "Name",
                    "variableSites": "Site",
                    "sequence": "Peptide", # We take this instead of sequenceVML, to match the other datasets' format
                    "accession_numbers": "Database_ID" # We take all accession numbers they have, instead of the singular accession_number column
                    })


                # Some rows have at least one localized acetylation site, but also have other acetylations that aren't localized. We'll drop those rows, if their localized sites are duplicated in another row, to avoid creating duplicates, because we only preserve information about the localized sites in a given row. However, if the localized sites aren't duplicated in another row, we'll keep the row.
                localization = df["accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA"].str.split("_", expand=True)
                unlocalized_to_drop = localization.index[~localization[3].eq(localization[4]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)] # Column 3 of the split localization data column is number of acetylation sites detected, and column 4 is number of acetylation sites localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                # Give it a multiindex
                df = df.set_index(["Name", "Site", "Peptide", "Database_ID"])
                cols_to_drop = [
                    "id",
                    "id.description",
                    "numColumnsVMsiteObserved",
                    "bestScore",
                    "bestDeltaForwardReverseScore",
                    "Best_scoreVML",
                    "sequenceVML",
                    "accessionNumber_VMsites_numVMsitesPresent_numVMsitesLocalizedBest_earliestVMsiteAA_latestVMsiteAA",
                    "protein_mw",
                    "species",
                    "speciesMulti",
                    "orfCategory",
                    "accession_number",
                    "protein_group_num",
                    "entry_name",
                    "GeneSymbol",
                    "VMsiteFlanks",
                    "Best_numActualVMSites_k",
                    "Best_numAmbiguousVMsites_k",
                    "Best_numLocalizedVMsites_k",
                    "StartAA"]

                df = df.drop(columns=cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                df.index.name="Patient_ID"
                self._data['ubiquitinomics'] = df

            elif file_name in ["lscc-v3.2-circular-rna-rsem-uq-log2.gct.gz", "lscc-v3.3-public-circular-rna-rsem-uq-log2.gct.gz"]:
                df = pd.read_csv(file_path, sep="\t", skiprows = 2, dtype=object)
                gene_filter = df['geneSymbol'] != 'na' #Filter out rows of metadata
                df = df[gene_filter]
                df = df.set_index("id")
                cols_to_drop = ["chrom","geneSymbol","end","start"]
                df = df.drop(columns = cols_to_drop)
                df = df.apply(pd.to_numeric)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data['circular_RNA'] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
        clinical = self._data["clinical"]
        clinical = clinical.reindex(master_index)

        # Impute any NaNs in Sample_Tumor_Normal
        clinical["Sample_Tumor_Normal"] = clinical["Sample_Tumor_Normal"].where(cond=~(pd.isnull(clinical["Sample_Tumor_Normal"]) & clinical.index.str.endswith(".N")), other="Normal")
        clinical["Sample_Tumor_Normal"] = clinical["Sample_Tumor_Normal"].where(cond=~(pd.isnull(clinical["Sample_Tumor_Normal"]) & ~clinical.index.str.endswith(".N")), other="Tumor")

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = clinical

        # Replace periods with hyphens in all Patient_IDs
        for name in self._data.keys(): # Loop over just the keys to avoid any issues that would come if we looped over the values while editing them
            df = self._data[name]

            # Make the index a column so we can edit it
            df.index.name = "Patient_ID"
            df = df.reset_index()

            # Replace all '.' with '-'
            df["Patient_ID"] = df["Patient_ID"].str.replace(r"\.", "-", regex=True)
            df["Patient_ID"] = df["Patient_ID"].str.replace(r"-N$", ".N", regex=True) # If there's a "-N" at the end, it's part of the normal identifier, which we want to actually be ".N"

            # Set the index back to Patient_ID
            df = df.set_index("Patient_ID")

            self._data[name] = df

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='lung squamous cell carcinoma', pmid='', unpublished=True)


#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime
import logging
from gtfparse import read_gtf

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError
from cptac.utils import get_boxnote_text
from .mssmclinical import MssmClinical


class WashuHnscc(Dataset):

    def __init__(self, no_internet, version):
        """Load all of the bcmbrca dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """
        #ignore logging messages
        logger = logging.getLogger()
        logger.setLevel(logging.CRITICAL)
        
        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": [
                "HNSCC_discovery.dnp.annotated.exonic.maf.gz",
                "HNSCC_NAT_RNA-Seq_Expr_WashU_FPKM.tsv.gz",
                "HNSCC_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz",
                "HNSCC_mature_miRNA_combined.tsv",
                "HNSCC_precursor_miRNA_combined.tsv",
                "HNSCC_total_miRNA_combined.tsv",
                "HNSCC_xCell.txt",
                "CIBERSORT.Output_Abs_HNSCC.txt",
                "HNSCC.gene_level.from_seg.filtered.tsv",
                "gencode.v22.annotation.gtf.gz",
                "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz",
                "README_miRNA",
                "README_CIBERSORT",
                "README_xCell",
                "README_somatic_mutation_WXS",
                "README_gene_expression",
                "README.boxnote",
                "README_ESTIMATE_WashU"
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="washuhnscc", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)
        
        # get clinical df (used to slice out cancer specific patient_IDs in tumor_purity file)
        mssmclin = MssmClinical(no_internet=no_internet, version=version, filter_type='pancanhnscc') #_get_version - pancandataset
        clinical_df = mssmclin.get_clinical()
        

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below

            

            if file_name == "HNSCC_discovery.dnp.annotated.exonic.maf.gz": # Note that we use the "file_name" variable to identify files. That way we don't have to use the whole path.
                df = pd.read_csv(file_path, sep='\t')    
                # Rename the columns we want to keep to the appropriate names
                df = pd.read_csv(file_path, sep='\t')    
                df['Patient_ID'] = df.loc[:, 'Tumor_Sample_Barcode']
                df = df.rename(columns={
                         "Hugo_Symbol":"Gene",
                         "Gene":"Gene_Database_ID",
                         "Variant_Classification":"Mutation",
                         "HGVSp_Short":"Location"})

                df = df.set_index("Patient_ID")
                df = df[ ['Gene'] + ["Mutation"] + ["Location"] + [ col for col in df.columns if col not in ["Gene","Mutation","Location"] ] ]
                df.index = df.index.str.replace(r"_T", "", regex=True)     
                self._data["somatic_mutation"] = df
              
            if file_name == "HNSCC_NAT_RNA-Seq_Expr_WashU_FPKM.tsv.gz":
                df_norm = pd.read_csv(file_path, sep='\t')
                #change names to universal package names
                df_norm = df_norm.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})  
                df_norm = df_norm.set_index(["Name", "Database_ID"])
                df_norm = df_norm.sort_index()
                df_norm = df_norm.T #transpose
                df_norm.index.name = "Patient_ID"
                df_norm.index = df_norm.index.str.replace(r"-A", ".N", regex=True) #remove label for tumor samples
                self._helper_tables["transcriptomics_norm"] = df_norm
                    
            if file_name == "HNSCC_tumor_RNA-Seq_Expr_WashU_FPKM.tsv.gz":
                df = pd.read_csv(file_path, sep='\t')
                #change names to universal package names
                df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                df = df.set_index(["Name", "Database_ID"])
                df = df.sort_index()
                df = df.T #transpose 
                df.index.name = "Patient_ID"
                df.index = df.index.str.replace(r"-T", "", regex=True) #remove label for tumor samples
                self._helper_tables["transcriptomics_tumor"] = df
                                        
            elif 'miRNA_combined' in file_name:
                miRNA_type = file_name.split('_')[1] # get type of miRNA data (precursor, mature, or total)
                if miRNA_type == 'mature':
                    df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias', 'Derives_from'])
                else:
                    df = pd.read_csv(file_path, delimiter = '\t', index_col = ['Name', 'ID','Alias'])
                df = df.transpose()
                df = average_replicates(df, common = '\.\d$') # average duplicates for C3L-02617 and C3N-02727
                df.index = df.index.str.replace('\.T$','', regex = True)
                df.index = df.index.str.replace('\.A$','.N', regex = True)
                df.index.name = 'Patient_ID'                
                # Sort
                normal = df.loc[df.index.str.contains('\.N$', regex =True)]
                normal = normal.sort_values(by=["Patient_ID"])
                tumor = df.loc[~ df.index.str.contains('\.N$', regex =True)]
                tumor = tumor.sort_values(by=["Patient_ID"])
                all_df = tumor.append(normal)
                self._data[miRNA_type+'_miRNA'] = all_df
                
            elif file_name == "HNSCC_xCell.txt":
                df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
                df = df.transpose()
                df.columns.name = 'Name'
                df.index.name = 'Patient_ID'
                df.index = df.index.str.replace(r'-T$', '', regex=True) # remove label for tumor samples
                df.index = df.index.str.replace(r'-A$', '.N', regex=True) # change label for normal samples
                self._data["xcell"] = df
                
            elif file_name == "CIBERSORT.Output_Abs_HNSCC.txt":
                df = pd.read_csv(file_path, sep = '\t', index_col = 0) 
                df.index.name = 'Patient_ID'
                df.columns.name = 'Name'
                df.index = df.index.str.replace(r'-T$', '', regex=True) 
                df.index = df.index.str.replace(r'-A$', '.N', regex=True)
                self._data["cibersort"] = df
                
            elif file_name == "HNSCC.gene_level.from_seg.filtered.tsv":
                df = pd.read_csv(file_path, sep="\t")
                df = df.rename(columns={"Gene": "Name"})
                df = df.set_index("Name")
                self._data["CNV"] = df
                
            elif file_name == "gencode.v22.annotation.gtf.gz":
                df = read_gtf(file_path)
                df = df[["gene_name","gene_id"]]
                df = df.drop_duplicates()
                df = df.rename(columns={"gene_name": "Name","gene_id": "Database_ID"})
                df = df.set_index("Name")
                self._helper_tables["CNV_gene_ids"] = df
                
            elif file_name == "CPTAC_pancan_RNA_tumor_purity_ESTIMATE_WashU.tsv.gz":
                df = pd.read_csv(file_path, sep = "\t", na_values = 'NA')
                df.Sample_ID = df.Sample_ID.str.replace(r'-T', '', regex=True) # only tumor samples in file
                df = df.set_index('Sample_ID') 
                df.index.name = 'Patient_ID' 
                # Use list of patient_ids to slice out cancers                
                patient_ids = clinical_df.index.to_list()
                df = df.loc[df.index.isin(patient_ids)]                
                self._data["tumor_purity"] = df
                                
            elif file_name == "README_miRNA":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_miRNA"] = reader.read()
                    
            elif file_name == "README_CIBERSORT":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_cibersort"] = reader.read()
                    
            elif file_name == "README_xCell":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_xcell"] = reader.read()
            
            elif file_name == "README_somatic_mutation_WXS":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_somatic_mutation"] = reader.read()
                    
            elif file_name == "README_gene_expression":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_transcriptomics"] = reader.read()
               
            elif file_name == "README.boxnote":
                self._readme_files["readme_cnv"] = get_boxnote_text(file_path)
             
            elif file_name == "README_ESTIMATE_WashU":
                with open(file_path, 'r') as reader:
                    self._readme_files["readme_tumor_purity"] = reader.read()
                    
        
        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')

        # combine and create transcriptomic dataframe            
        rna_tumor = self._helper_tables.get("transcriptomics_tumor")
        rna_normal = self._helper_tables.get("transcriptomics_norm") # Normal entries are already marked with 'N' on the end of the ID
        rna_combined = rna_tumor.append(rna_normal)
        self._data["transcriptomics"] = rna_combined

        
        # CNV
        cnv = self._data["CNV"]
        gene_ids = self._helper_tables["CNV_gene_ids"]
        df = cnv.join(gene_ids,how = "left") #merge in gene_ids 
        df = df.reset_index()
        df = df.set_index(["Name", "Database_ID"]) #create multi-index
        df = df.T
        df.index.name = 'Patient_ID'
        self._data["CNV"] = df 
        
        self._data = sort_all_rows_pancan(self._data) # Sort IDs (tumor first then normal)

        
        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN.
#        new_clinical = self._data["clinical"]
#        new_clinical = new_clinical.reindex(master_index)

        # Add a column called Sample_Tumor_Normal to the clinical dataframe indicating whether each sample was a tumor or normal sample. Use a function from dataframe_tools to generate it.

        ###FILL: Your dataset should have some way that it marks the Patient IDs
        ### of normal samples. The example code below is for a dataset that
        ### marks them by putting an 'N' at the beginning of each one. You will
        ### need to write a lambda function that takes a given Patient_ID string
        ### and returns a bool indicating whether it corresponds to a normal
        ### sample. Pass that lambda function to the 'normal_test' parameter of
        ### the  generate_sample_status_col function when you call it. See 
        ### cptac/dataframe_tools.py for further function documentation.
        ###START EXAMPLE CODE###################################################
#        sample_status_col = generate_sample_status_col(new_clinical, normal_test=lambda sample: sample[0] == 'N')
        ###END EXAMPLE CODE#####################################################

#        new_clinical.insert(0, "Sample_Tumor_Normal", sample_status_col)

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
#        self._data['clinical'] = new_clinical

        # Edit the format of the Patient_IDs to have normal samples marked the same way as in other datasets. 
        
        ###FILL: You will need to pass the proper parameters to correctly
        ### reformat the patient IDs in your dataset. The standard format is to
        ### have the string '.N' appended to the end of the normal patient IDs,
        ### e.g. the  normal patient ID corresponding to C3L-00378 would be
        ### C3L-00378.N (this way we can easily match two samples from the same
        ### patient). The example code below is for a dataset where all the
        ### normal samples have  an "N" prepended to the patient IDs. The
        ### reformat_normal_patient_ids function erases that and puts a ".N" at
        ### the end. See cptac/dataframe_tools.py for further function
        ### documentation.
        ###START EXAMPLE CODE###################################################
#        self._data = reformat_normal_patient_ids(self._data, existing_identifier="N", existing_identifier_location="start")
        ###END EXAMPLE CODE#####################################################

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
#        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
#        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .pancandataset import PancanDataset
from ..dataframe_tools import sort_all_rows_pancan

from .mssmclinical import MssmClinical
from .pdcucec import PdcUcec
from .umichucec import UmichUcec
from .washuucec import WashuUcec
from .bcmucec import BcmUcec
from .broaducec import BroadUcec
from .harmonized import Harmonized
from .joiningdataset import JoiningDataset

#List Sources to be downloaded
SOURCES = [
    "bcmucec",
    "mssmclinical",
    "pdcucec",
    "umichucec",
    "washuucec",
    "broaducec",
    "harmonized"
    
]

class PancanUcec(PancanDataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all the data sources with UCEC data and provide an interface to them."""

        super().__init__(cancer_type="pancanucec", version=version, no_internet=no_internet)

        self._datasets["bcm"] = BcmUcec(no_internet=no_internet, version=self._get_version("bcm"))
        self._datasets["broad"] = BroadUcec(no_internet=no_internet, version=self._get_version("broad"))
        self._datasets["mssm"] = MssmClinical(no_internet=no_internet, version=self._get_version("mssm"), filter_type='pancanucec')
        self._datasets["pdc"] = PdcUcec(no_internet=no_internet, version=self._get_version("pdc"))
        self._datasets["umich"] = UmichUcec(no_internet=no_internet, version=self._get_version("umich"))
        self._datasets["washu"] = WashuUcec(no_internet=no_internet, version=self._get_version("washu"))
        self._datasets["harmonized"] = Harmonized(no_internet=no_internet, version=self._get_version("harmonized"), filter_type='pancanucec')
        
        join_dict = {k: v._data for k, v in self._datasets.items()}
        self._joining_dataset = JoiningDataset(join_dict)
        
        self._pancan_unionize_indices()



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
import datetime

from cptac.dataset import Dataset
from cptac.dataframe_tools import *
from cptac.exceptions import FailedReindexWarning, PublicationEmbargoWarning, ReindexMapError
from cptac.utils import get_boxnote_text


class UmichBrca(Dataset):

    def __init__(self, no_internet, version):
        """Load all of the umichbrca dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.
        valid_versions = ["1.0"]

        data_files = {
            "1.0": ["Report_abundance_groupby=protein_protNorm=MD_gu=2.tsv",
                    "Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv",
                    "prosp-brca-all-samples.txt",
                    "README_v3.boxnote", # proteomics 
                    "README.boxnote" # phosphoproteomics 
            ]
        }

        # Call the parent class __init__ function
        super().__init__(cancer_type="umichbrca", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths: # Loops through files variable

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file. We'll use this to identify files for parsing in the if/elif statements below
                
            #Proteomics
            if file_name == "Report_abundance_groupby=protein_protNorm=MD_gu=2.tsv":
                df = pd.read_csv(file_path, sep = "\t") 
                df['Database_ID'] = df.Index.apply(lambda x: x.split('|')[0]) # get protein identifier 
                df['Name'] = df.Index.apply(lambda x: x.split('|')[6]) # get protein name 
                df = df.set_index(['Name', 'Database_ID']) # set multiindex
                df = df.drop(columns = ['Index', 'MaxPepProb', 'NumberPSM', 'Gene']) # drop unnecessary  columns
                df = df.transpose()
                ref_intensities = df.loc["ReferenceIntensity"] # get reference intensities to use to calculate ratios 
                df = df.subtract(ref_intensities, axis="columns") # subtract reference intensities from all the values
                df = df.iloc[1:,:] # drop ReferenceIntensity row 
                df.index.name = 'Patient_ID'
                
                # drop ending of CPT retrospective samples to match cptac 
                df = df.rename(index={'CPT0008140004':'CPT000814', 'CPT0018460005': 'CPT001846', 
                                      '604':'CPT000814'}) # 604 mapped to CPT000814 in the proteomics file from the PDC pipeline 
    
                drop_cols = ['RetroIR', 'RetroIR.1',
                   'RefInt_Pool01', 'RefInt_Pool02', 'RefInt_Pool03', 'RefInt_Pool04',
                   'RefInt_Pool05', 'RefInt_Pool06', 'RefInt_Pool07', 'RefInt_Pool08',
                   'RefInt_Pool09', 'RefInt_Pool10', 'RefInt_Pool11', 'RefInt_Pool12',
                   'RefInt_Pool13', 'RefInt_Pool14', 'RefInt_Pool15', 'RefInt_Pool16',
                   'RefInt_Pool17']                
                df = df.drop(drop_cols, axis = 'index') # drop quality control and ref intensity cols
                self._data["proteomics"] = df
                
            #Phosphoproteomics    
            elif file_name == "Report_abundance_groupby=multi-site_protNorm=MD_gu=2.tsv":
                df = pd.read_csv(file_path, sep = "\t") 
                 # Parse a few columns out of the "Index" column that we'll need for our multiindex
                df[['Database_ID','Transcript_ID',"Gene_ID","Havana_gene","Havana_transcript","Transcript","Name","Site"]] = df.Index.str.split("\\|",expand=True)
                df[['num1','start',"end","detected_phos","localized_phos","Site"]] = df.Site.str.split("_",expand=True) 

                # Some rows have at least one localized phosphorylation site, but also have other 
                # phosphorylations that aren't localized. We'll drop those rows, if their localized 
                # sites are duplicated in another row, to avoid creating duplicates, because we only 
                # preserve information about the localized sites in a given row. However, if the localized 
                # sites aren't duplicated in another row, we'll keep the row.
                unlocalized_to_drop = df.index[~df["detected_phos"].eq(df["localized_phos"]) & df.duplicated(["Name", "Site", "Peptide", "Database_ID"], keep=False)]# dectected_phos of the split "Index" column is number of phosphorylations detected, and localized_phos is number of phosphorylations localized, so if the two values aren't equal, the row has at least one unlocalized site
                df = df.drop(index=unlocalized_to_drop)

                df = df[df['Site'].notna()] # only keep columns with phospho site 
                df = df.set_index(['Name', 'Site', 'Peptide', 'Database_ID']) # Create a multiindex in this order.
                #drop columns not needed in df 
                df.drop(['Gene', "Index", "num1", "start", "end", "detected_phos", "localized_phos", "Havana_gene", 
                         "Havana_transcript", "MaxPepProb", "Gene_ID", "Transcript_ID", "Transcript"], axis=1, inplace=True)
                df = df.T #transpose df 
                ref_intensities = df.loc["ReferenceIntensity"]# Get reference intensities to use to calculate ratios 
                df = df.subtract(ref_intensities, axis="columns") # Subtract ref intensities from all the values, to get ratios
                df = df.iloc[1:,:] # drop ReferenceIntensity row 
                # drop ending of CPT retrospective samples to match cptac
                df = df.rename(index={'CPT0008140004':'CPT000814', 'CPT0018460005': 'CPT001846', 
                                      '604':'CPT000814'}) # 604 mapped to CPT000814 in pdc index
                
                drop_cols = ['RetroIR','RetroIR.1','RefInt_Pool01-1','RefInt_Pool02-1',
                             'RefInt_Pool03-1','RefInt_Pool04-1','RefInt_Pool05-1','RefInt_Pool06-1',
                             'RefInt_Pool07-1','RefInt_Pool08-1','RefInt_Pool09-1','RefInt_Pool10-1',
                             'RefInt_Pool11-1','RefInt_Pool12-1','RefInt_Pool13-1','RefInt_Pool14-1',
                             'RefInt_Pool15-1','RefInt_Pool16-1','RefInt_Pool17-1']
                # Drop quality control and ref intensity cols
                df = df.drop(drop_cols, axis = 'index')
                self._data["phosphoproteomics"] = df
            
            
            # prosp-brca-all-samples.txt shows which patient IDs have normal samples and which have replicates 
            # This file can be found on Box under CPTAC/cptac/pancan/helper_files
            if file_name == "prosp-brca-all-samples.txt":
                df = pd.read_csv(file_path, sep = "\t")
                df = df[['Participant', 'id', 'Type']]
                self._helper_tables["map_ids"] = df
                
            elif file_name == "README_v3.boxnote":
                self._readme_files["readme_proteomics"] = get_boxnote_text(file_path)
                
            elif file_name == "README.boxnote":
                self._readme_files["readme_phosphoproteomics"] = get_boxnote_text(file_path)

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = f"Formatting {self.get_cancer_type()} dataframes..."
        print(formatting_msg, end='\r')
        
        # Get patient IDs with normal samples or replicates (from mapping file)         
        # 7 IDs with replicates: '11BR031', '11BR053', '11BR036', '11BR060', '14BR005', '11BR011', '21BR010'
        # 18 IDs with normals: '11BR074', '11BR073', '20BR007', '21BR010', '11BR017', '05BR029', '18BR003', '11BR030',
        #   '01BR027','11BR025', '11BR047', '11BR028', '11BR020', '20BR008', '11BR024', '11BR023', '11BR015', '11BR006'
        map_df = self._helper_tables["map_ids"]
        # Get IDs with replicates 
        replicate_list = list(set(map_df.loc[map_df.id.str.contains('REP')].Participant))
        replicate_list.remove('RetroIR') 
        replicate_list = [x[1:] for x in replicate_list]
        # Get IDs with normals 
        norm_df = map_df.loc[map_df.Type == 'Adjacent_Normal'] # get all patient_IDs with normal samples
        norm_df.index = norm_df.Participant.apply(lambda x: x[1:]+'.1') #remove initial 'X' and add '.1' (did not correlate well)
        not_tumor = norm_df.index.to_list()
        
        # Drop samples that don't correlate well with the original cptac tumor values
        # The proteomics and phosphoproteomics file contained duplicate patient IDs with unique values. 18 of these were IDs 
        # with both tumor and normal samples (according to the mapping file). To find which ID contained values nearest to the 
        # original cptac tumor measurements, we calculated pearson correlations and created scatterplots to compare 
        # flagship values to each of the duplicate patient ID values. The first occurrance of every patient ID in the 
        # file with a normal sample showed a high correlation with the flagship values (average around 0.9). 
        # The second occurrence showed a low correlation with the flagship tumor values (average around  0.2). 
        # In Brca, normal samples were dropped in downstream analysis because of quality control issues. Therefore, we dropped 
        # the second occurrence of the 18 patient IDs with a normal sample because they did not correlate well with the 
        # flagship tumor values and are likely normal samples. There were 7 IDs with replicates shown in the mapping file 
        # (prosp-brca-all-samples.txt). The same method was used to check that these correlated well with their respective 
        # flagship cptac tumor values. Replicates were averaged (consistent with the handling of other replicates in the pancan 
        # module).
        # note:  21BR010.1 had a correlation of 0.275 (so dropped), and 21BR010.2 had correlation of 0.848 (so averaged)        
        # A file containing the correlations can be downloaded at: 
        # https://byu.box.com/shared/static/jzsq69bd079oq0zbicw4w616hyicd5ev.xlsx
        
        if self._version == "1.0":
            # Proteomics
            prot = self._data["proteomics"]
            prot = prot.loc[ ~ prot.index.isin(not_tumor)] # drop rows that don't correlate well with respective cptac tumor 
            prot = average_replicates(prot, replicate_list) # average 7 IDs with replicates  
            self._data["proteomics"] = prot

            # Phosphoproteomics
            phos = self._data["phosphoproteomics"]
            phos = phos.loc[ ~ phos.index.isin(not_tumor)] # drop rows that don't correlate well with respective cptac tumor 
            phos = average_replicates(phos, replicate_list) # average 7 IDs with replicates
            self._data["phosphoproteomics"] = phos
        
        
        # Sort rows (tumor first then normal) and columns by first level (protein/gene name)
        self._data = sort_all_rows_pancan(self._data)  
        

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .pancandataset import PancanDataset

from .mssmclinical import MssmClinical
from .pdcpdac import PdcPdac
from .umichpdac import UmichPdac
from .washupdac import WashuPdac
from .bcmpdac import BcmPdac
from .broadpdac import BroadPdac
from .harmonized import Harmonized
from .joiningdataset import JoiningDataset

#List Sources to be downloaded
SOURCES = [
    "bcmpdac",
    "mssmclinical",
    "pdcpdac",
    "umichpdac",
    "washupdac",
    "broadpdac",
    "harmonized"
    
]

class PancanPdac(PancanDataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all the data sources with PDAC data and provide an interface to them."""

        super().__init__(cancer_type="pancanpdac", version=version, no_internet=no_internet)

        self._datasets["bcm"] = BcmPdac(no_internet=no_internet, version=self._get_version("bcm"))
        self._datasets["broad"] = BroadPdac(no_internet=no_internet, version=self._get_version("broad"))
        self._datasets["mssm"] = MssmClinical(no_internet=no_internet, version=self._get_version("mssm"), filter_type='pancanpdac')
        self._datasets["pdc"] = PdcPdac(no_internet=no_internet, version=self._get_version("pdc"))
        self._datasets["umich"] = UmichPdac(no_internet=no_internet, version=self._get_version("umich"))
        self._datasets["washu"] = WashuPdac(no_internet=no_internet, version=self._get_version("washu"))
        self._datasets["harmonized"] = Harmonized(no_internet=no_internet, version=self._get_version("harmonized"), filter_type='pancanpdac')
        
        join_dict = {k: v._data for k, v in self._datasets.items()}
        self._joining_dataset = JoiningDataset(join_dict)
        
        self._pancan_unionize_indices()



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import warnings
from .exceptions import CptacDevError, ReindexMapError, FailedReindexWarning
from contextlib import contextmanager
import sys, os
import mygene


@contextmanager
def suppress_stdout():
    """Temporarily stops output from being printed. Use before a function when you don't want its output to
    be printed. 
    Example: with suppress_stdout(): 
                 function()
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout
            

def map_database_to_gene_pdc(df, database_name = 'refseq', sep = ':'): 
    """Add gene name to dataframe based on database IDs. This function is based on the format of PDC files. 
    Only phosphoproteomics and acetylproteomics need the gene name added. 

    Parameters:
    df (dict): The dataframe that needs gene names.
    database_name (str): dataframe name to use with mygene 'scopes' parameter.
    sep (str): separator to split string with database ID. 

    Returns:
    df: The dataframe with gene name, database ID and site as a multiindex (sorted by gene name).
    """
    
    if isinstance(df.index, pd.core.indexes.multi.MultiIndex):
        was_multiindex = True
        df = df.reset_index() # make index not a multiindex
    else:
        was_multiindex = False
        
    df = df.T.reset_index()
    df[['Database_ID',"Site"]] = df.iloc[:, 0].str.split(sep, expand=True) # first column is reset index  
    df.Site = df.Site.str.upper() # capitalize all amino acids sites (for consistency)
    
    id_list = df.Database_ID.to_list()
    with suppress_stdout():
        mg = mygene.MyGeneInfo()
        db_results = mg.querymany(id_list, scopes=database_name, fields='symbol', species='human') # map ID to gene
    db_to_gene = {results['query']: results['symbol'] for results in db_results \
                           if 'notfound' not in results.keys()} # get mapping dictionary of ID and gene name
    df['Name'] = df['Database_ID'].replace(db_to_gene) # add gene name 
    df = df.set_index(['Name','Site','Database_ID']) # set multiindex
    df = df.sort_index(level='Name', axis = 'index') # sort based on gene name
    df = df.drop('index', axis = 1)
    df = df.T
    
    if was_multiindex == True:
        df = df.set_index([df.columns[-1], df.columns[-2]])
        df.index.names = ['case_submitter_id','aliquot_submitter_id']
        
    return df

def sort_all_rows_pancan(data_dict):
    """For all dataframes in the given dictionary, sort them first by sample status, 
    with tumor samples first, and then by the index. Also sorts columns by the first level. 

    Parameters:
    data_dict (dict): The dataframe dictionary of the dataset.

    Returns:
    dict: The dataframe dictionary, with the dataframes sorted by their indices. 
    The index is also given the standard name ('Patient_ID').
    Keys are str of dataframe names, values are pandas.DataFrame"""
    
    for name in data_dict.keys(): # Loop over the keys so we can alter the values without any issues
        df = data_dict[name]
        df = df.sort_index(axis = 'columns', level = 0) # sort columns based on first level
        if isinstance(df.index, pd.core.indexes.multi.MultiIndex):
            df.index.rename(['Patient_ID', 'Aliquot'], level = [0,1], inplace = True)
            new_df = df.sort_values('Patient_ID')
            data_dict[name] = new_df
        else:
            df.index.name = "Patient_ID" # set the name of the index
            # Sort normal samples
            normal = df.loc[df.index.str.contains('\.[NC]$', regex = True, na = False)]#'.N' for normal, '.C' for cored normals (in HNSCC)
            normal = normal.sort_index() # index should be Patient_ID
            # Sort tumor samples
            tumor = df.loc[~ df.index.str.contains('\.[NC]$', regex = True, na = False)]
            tumor = tumor.sort_index()
            # append normal to tumor
            all_df = tumor.append(normal)
            data_dict[name] = all_df 

    return data_dict


def rename_duplicate_labels(df, label_type='columns'):
    """Returns a df with unique labels for columns or indices 
    Parameters:
    df (pandas.DataFrame): The df containing duplicate labels.
    label_type (str): use 'columns' to make unique column labels and 'index' to make unique indices. 
    
    Returns:
    pandas.DataFrame: df with with unique labels. ".i" is appended to every duplicate, increasing incrementally.
    """
    if label_type == 'columns':
        labs = pd.Series(df.columns[:])

        for dup in labs[labs.duplicated()].unique(): 
            labs[labs[labs == dup].index.values.tolist()] = [dup + '.' + str(i) if i != 0 else dup for i in range(sum(labs == dup))]

        # rename the columns with the labs list.
        df.columns=labs
        
    elif label_type == 'index':
        labs = pd.Series(df.index[:])

        for dup in labs[labs.duplicated()].unique(): 
            labs[labs[labs == dup].index.values.tolist()] = [dup + '.' + str(i) if i != 0 else dup for i in range(sum(labs == dup))]

        # rename the columns with the labs list.
        df.index=labs
    return df

def average_replicates(df, id_list = [], normal_identifier = '.N', common = '\.', to_drop = '\.\d$'):
    """Returns a df with one row for each patient_ID (all replicates for a patient are averaged)
    Parameters:
    df (pandas.DataFrame): The df containing replicates (duplicate entries for the same tissue_type).
    id_list: list of IDs with replicates (use the ID format that is common between replicates so the 
            list can be used to slice out all replicates for that ID). Make sure the IDs 
            in the list include the symbol that distinguishes normal samples ('.N' or '-N').
    common: regex string that is common between replicates (identifies duplicate entries)
    to_drop: regex string to drop to find each patient_ID that has replicates (used to slice out all replicates)
    
    Returns:
    pandas.DataFrame: df with with replicate rows averaged and one row for each patient_ID.
    """
    # If no list of replicate IDs is given, make list from common regex 
    if len(id_list) == 0:
        replicate_df = df[df.index.str.contains(common)]
        patient_ids = pd.Series(replicate_df.index) # create series of replicate IDs to prep removing appended ".i"
        ids = patient_ids.replace(to_drop, '', regex=True)
        id_list = list(set(ids)) #id_list contains only patient_IDs of replicates (without #s)

    new_df = df.copy()
    for patient_ID in id_list:
        # Can slice only normals with patient_ID because of normal identifier (won't slice out tumors)
        if normal_identifier in patient_ID:
            id_df = df[df.index.str.contains(patient_ID, regex = True)] # slice out replicates for a single patient
        # If tumor, need to slice out normals
        else:
            if normal_identifier == '.N':
                norm_regex = '\.N' # prep for regex use
            else:
                norm_regex = normal_identifier
            id_df = df[df.index.str.contains(patient_ID, regex = True) & \
                       ~ df.index.str.contains(norm_regex, regex = True)] # don't include normals
        #print(id_df.index.to_list())
        vals = list(id_df.mean(axis=0)) 
        new_df = new_df.drop(id_df.index.to_list(), axis = 'index') # drop unaveraged rows
        new_df.loc[patient_ID] = vals # add averaged row

    return new_df

def unionize_indices(dataset, exclude=[]):
    """Return a union of all indices in a dataset, without duplicates.

    Parameters:
    dataset (dict of str: pandas.DataFrame): The data dictionary containing the dataset.
    exclude (str or list of str, optional): A list of dataframes to exclude when unionizing indices.

    Returns:
    pandas.Index: Union of all indices in the dataset, without duplicates.
    """
    if isinstance(exclude, str): # If it's a single dataframe name, make it a list so we can treat everything the same
        exclude = [exclude]
    indices = [df.index for name, df in dataset.items() if name not in exclude]
    master_index = pd.Index([])
    for index in indices:
        master_index = master_index.union(index)
        master_index = master_index.drop_duplicates()
   
    return master_index

def generate_sample_status_col(df, normal_test):
    """Create a sample status column, called Sample_Tumor_Normal, for a dataframe.

    Parameters:
    df (pandas.DataFrame): The dataframe to create a Sample_Status column for, indexed with Patient_IDs.
    normal_test (function): A function that takes a given Patient_ID and returns a bool indicating whether it corresponds to a normal sample.

    Returns:
    pandas Series: A sample status column for the dataframe.
    """
    sample_status_array = np.where(df.index.map(normal_test), "Normal", "Tumor")
    sample_status_col = pd.Series(data=sample_status_array, index=df.index.copy())
    return sample_status_col

def get_reindex_map(series):
    """Generate a reindexing map from a series where the index is the new indices, and the values are the old indices.

    Parameters:
    series (pandas Series): The series to generate the reindex map from.

    Returns:
    pandas Series: The reindexing map, with the old inidicies as the index, and the new indices as the values.
    """
    old_index_name = series.name
    new_index_name = series.index.name
    if new_index_name is None:
        new_index_name = "index" # This is the default name pandas gives unnamed indices when they're made columns

    # Check that the mapping is one to one
    series = series.dropna()
    if len(series) != len(series.drop_duplicates()):
        raise ReindexMapError("Reindex map is not one to one.")

    # Make the index the values, and the values the index
    df = series.reset_index()
    df = df.set_index(old_index_name)
    reindex_map = df[new_index_name]
    return reindex_map

def reindex_dataframe(df, reindex_map, new_index_name, keep_old):
    """Reindex a dataframe based on a mapping of the old index values to new ones.

    Parameters:
    df (pandas.DataFrame): The dataframe to reindex.
    reindex_map (dict or pandas Series): A dictionary or pandas Series with the old index values as the keys or index, and the new ones as the values.
    new_index_name (str): The desired name for the new index.
    keep_old (bool): Whether to retain the old index in the dataframe as a column.

    Returns:
    pandas.DataFrame: A copy of the given dataframe, with the new index.
    """
    if not df.index.isin(reindex_map.keys()).all(): # This works for a dict or a pandas Series, because Series have a .keys() attribute that's an alias for the index
        not_in = df.index[~(df.index.isin(reindex_map.keys()))]
        raise ReindexMapError(not_in)

    new_index = df.index.map(reindex_map.get)

    if keep_old:
        df = df.reset_index()
    df.index = new_index
    df.index.name = new_index_name
    df = df.sort_index()
    return df

def reindex_all_sample_id_to_patient_id(data_dict, reindex_map, additional_to_keep_col=[], skip=[]):
    """Reindex all the dataframes with Patient_IDs instead of Sample_IDs

    Parameters:
    data_dict (keys are str, values are pandas.DataFrame): The data dictionary to reindex
    reindex_map (dict or pandas Series): A dictionary or pandas Series with the old index values (Sample_IDs) as the keys or index, and the new ones (Patient_IDs) as the values. Must map for all existing index values in the entire dataset.
    additional_to_keep_col (list of str, optional): The function will already keep the old index as a column in the clinical dataframe. If you want it to do this for any other dataframes, put their names in this list.
    skip (str or list of str, optional): A list of dataframes to skip when reindexing.

    Returns:
    dict: The data dictionary, with all dataframes reindexed with Patient_IDs
    """
    if isinstance(skip, str): # If it's a single dataframe name, make it a list so we can treat everything the same
        skip = [skip]

    dfs_to_delete = []
    dfs_to_keep_col = ["clinical"] + additional_to_keep_col

    for name in data_dict.keys(): # Only loop over keys, to avoid changing the structure of the object we're looping over

        # Skip any specified to skip
        if name in skip:
            continue

        df = data_dict[name]
        df.index.name = "Sample_ID" # So that it's labeled properly when we keep it as a column in the clinical dataframe.
        keep_old = name in dfs_to_keep_col # Keep the old Patient_ID index as a column in the clinical dataframe (and any additionally specified dataframes), so we have a record of it.

        try:
            df = reindex_dataframe(df, reindex_map, "Patient_ID", keep_old)
        except ReindexMapError:
            warnings.warn(f"Error reindexing {name} dataframe. At least one Sample_ID did not have corresponding Patient_ID mapped in clinical dataframe. {name} dataframe not loaded.", FailedReindexWarning, stacklevel=3) # stacklevel=3 ensures that the warning is registered as originating from the file that called the __init__ function, instead of from here directly, because the former is more useful information.
            dfs_to_delete.append(name)
            continue

        data_dict[name] = df

    for name in dfs_to_delete: # Delete any dataframes that had issues reindexing
        del data_dict[name]

    return data_dict

def reformat_normal_patient_ids(data_dict, existing_identifier=None, existing_identifier_location=None):
    """Reformat the patient IDs for normal samples to be marked by an appended ".N"

    Parameters:
    data_dict (dict): The data dictionary for a dataset. All the tables must be indexed by Patient IDs.
    existing_identifier (str, optional): A normal sample identifier that already exists on the normal samples' patient IDs, which we will remove before adding the new identifier. Default of None will cause nothing to be removed.
    existing_identifier_location (str, optional): Either "start" or "end": Indicates whether the existing identifier is at the beginning or end of the normal samples' patient IDs, so we know which end to remove it from. Optional if nothing is passed to the existing_identifier parameter.

    Returns:
    dict: The data dictionary for the dataset, with normal samples' patient IDs reformatted in the specified dataframes.
    """

    # Check parameters
    if (existing_identifier is None and existing_identifier_location is not None) or (existing_identifier is not None and existing_identifier_location is None):
        raise CptacDevError("Parameters existing_identifier and existing_identifier_location must either both be None, or both not be None.")

    sample_statuses_old_index = data_dict["clinical"]["Sample_Tumor_Normal"] # We'll need this every time, and we need it with the un-reformatted index

    for name in data_dict.keys(): # Loop over the keys so we can edit the values without any issues

        df = data_dict[name]

        # Add in the tumor/normal statuses for these samples, if they aren't already in the table
        added_sample_statuses = False # So we can keep track of whether to drop the column when we're done
        if "Sample_Tumor_Normal" not in df.columns:
            df = join_col_to_dataframe(df, sample_statuses_old_index)
            added_sample_statuses = True

        df.index.name = "Patient_ID" # So we can easily access it when we've made it into a column
        df = df.reset_index() # This makes the Patient_ID index a column, so we can edit it.

        if (existing_identifier is not None) and (existing_identifier_location is not None): # There's an existing normal sample identifier to remove
            existing_length = len(existing_identifier)

            if existing_identifier_location == "start":
                df["Patient_ID"] = df["Patient_ID"].where(
                    cond=(~((df["Sample_Tumor_Normal"] == "Normal") & (df["Patient_ID"].str[0:existing_length] == existing_identifier))),
                    other=df["Patient_ID"].str[existing_length:]
                )

            elif existing_identifier_location == "end":
                df["Patient_ID"] = df["Patient_ID"].where(
                    cond=(~((df["Sample_Tumor_Normal"] == "Normal") & (df["Patient_ID"].str[-existing_length:] == existing_identifier))),
                    other=df["Patient_ID"].str[:-existing_length] # Note that we use the negative of the existing length, since we're working with the end of the string
                )

            else:
                raise CptacDevError("existing_identifier_location parameter must be either 'start' or 'end'")

        # Append ".N" to the patient IDs of normal samples
        df["Patient_ID"] = df["Patient_ID"].where(
            cond=(~(df["Sample_Tumor_Normal"] == "Normal")),
            other=df["Patient_ID"] + ".N"
        )

        # Set the index to the reformatted Patient IDs
        df = df.set_index("Patient_ID")

        # If we added the Sample_Tumor_Normal column, drop it
        if added_sample_statuses:
            if isinstance(df.columns, pd.MultiIndex):
                df = df.drop(columns="Sample_Tumor_Normal", level=0) # level=0 prevents a PerformanceWarning
            else:
                df = df.drop(columns="Sample_Tumor_Normal")

        # Put the dataframe with reformatted patient IDs back into the data dictionary
        data_dict[name] = df

    return data_dict

def join_col_to_dataframe(df, col):
    """Join a sample status column into a dataframe, automatically accounting for whether the dataframe has a column multiindex or not.

    Parameters:
    df (pandas.DataFrame): The dataframe to join the column into
    col (pandas Series): The column to join into the dataframe, with a matching index.

    Returns:
    pandas.DataFrame: The dataframe with the column joined in.
    """
    col_df = col.to_frame().copy(deep=True)
    
    # Make sure the columns axes all have the same name
    df.columns.name = "Name"
    col_df.columns.name = "Name"

    # If df has a column multiindex, edit the col_df column index to match, so we can join them
    if col_df.columns.names != df.columns.names:
        col_df.columns = add_index_levels(to=col_df.columns, source=df.columns) 
    
    if col_df.columns.names != df.columns.names: # Just to make sure
        raise CptacDevError(f"col_df's column axes had levels not found in the dataframe's columns.")

    df = df.join(col_df, how="left") # We do a left join because we only want rows that exist in our dataframe

    return df

def standardize_axes_names(data_dict):
    """For all dataframes in the given dictionary, sets the name of the index axes to "Patient_ID", because that's what they all are by that point, and sets the name of the column axes to "Name".

    Parameters:
    data_dict (dict): The dataframe dictionary of the dataset.

    Returns:
    dict: The dataframe dictionary, with the dataframe axes' names standardized. Keys are str of dataframe names, values are pandas.DataFrame
    """
    for name in data_dict.keys(): # Loop over the keys so we can alter the values without any issues
        df = data_dict[name]
        df.index.name = "Patient_ID"
        df.columns.name = "Name"
        data_dict[name] = df

    return data_dict

def sort_all_rows(data_dict):
    """For all dataframes in the given dictionary, sort them first by sample status, with tumor samples first, and then by the index.

    Parameters:
    data_dict (dict): The dataframe dictionary of the dataset.

    Returns:
    dict: The dataframe dictionary, with the dataframes sorted by their indices. Keys are str of dataframe names, values are pandas.DataFrame
    """
    # Get the Sample_Tumor_Normal column as a single-column dataframe
    sample_status_col = data_dict["clinical"]["Sample_Tumor_Normal"].copy(deep=True) # We'll need this every time

    for name in data_dict.keys(): # Loop over the keys so we can alter the values without any issues
        df = data_dict[name]
        df = sort_df_by_sample_status(df, sample_status_col)
        data_dict[name] = df

    return data_dict

def sort_df_by_sample_status(df, sample_status_col):
    """Sort a dataframe first by sample status, with tumor first, and then by Patient_ID.

    df (pandas.DataFrame): The dataframe to sort.
    sample_status_col (pandas Series): The Sample_Tumor_Normal column for the dataset.

    Returns:
    pandas.DataFrame: The dataframe, sorted.
    """
    # Add in the tumor/normal statuses for these samples, if they aren't already in the table
    added_sample_statuses = False # So we can keep track of whether to drop the column when we're done

    if "Sample_Tumor_Normal" not in df.columns:
        df = join_col_to_dataframe(df, sample_status_col)
        added_sample_statuses = True

    # Sort first by the Sample_Tumor_Normal column, and then by the index
    df.index.name = "Patient_ID" # To make sure we can reference it in the next line
    df = df.sort_values(by=["Sample_Tumor_Normal", "Patient_ID"], ascending=[False, True]) # Sorts first by sample status, and in descending order, so "Tumor" samples are first

    # If we added the Sample_Tumor_Normal column, drop it
    if added_sample_statuses:
        if isinstance(df.columns, pd.MultiIndex):
            df = df.drop(columns="Sample_Tumor_Normal", level=0) # level=0 prevents a PerformanceWarning
        else:
            df = df.drop(columns="Sample_Tumor_Normal")

    return df

def add_index_levels(to, source, fill=""):
    """Add levels to the "to" index so it has all levels in the "source" index. The possible levels are, in this order: "Name", "Site", "Peptide", "Database_ID"

    Parameters:
    to (pandas.Index or pandas.MultiIndex): The index to add levels to.
    source (pandas.Index or pandas.MultiIndex): The index to match the levels of.
    fill (optional): Value to fill empty levels with. Default is an empty string, which allows us to select a column with just the first level. This is useful for boolean filters.

    Returns:
    pandas.MultiIndex: The levels of "to", with any levels from "source" that "to" didn't have originally.
    """
    to_set = set(to.names)
    source_set = set(source.names)
    if source_set <= to_set:
        return to # Because otherwise we'd just end up constructing a duplicate of "to", and who would want to do that?

    all_names = ["Name", "Site", "Peptide", "Database_ID"]
    levels = {}

    for name in all_names:
        if name in to.names:
            levels[name] = to.get_level_values(name)
        elif name in source.names:
            levels[name] = [fill for i in range(to.size)]

    new_columns = pd.MultiIndex.from_arrays(list(levels.values()), names=list(levels.keys()))
    return new_columns



#   Copyright 2018 Samuel Payne sam_payne@byu.edu
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#       http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd
import numpy as np
import os
import warnings
from .dataset import Dataset
from .file_download import update_index
from .dataframe_tools import *
from .exceptions import FailedReindexWarning, ReindexMapError

class Ovarian(Dataset):

    def __init__(self, version="latest", no_internet=False):
        """Load all of the ovarian dataframes as values in the self._data dict variable, with names as keys, and format them properly.

        Parameters:
        version (str, optional): The version number to load, or the string "latest" to just load the latest building. Default is "latest".
        no_internet (bool, optional): Whether to skip the index update step because it requires an internet connection. This will be skipped automatically if there is no internet at all, but you may want to manually skip it if you have a spotty internet connection. Default is False.
        """

        # Set some needed variables, and pass them to the parent Dataset class __init__ function

        valid_versions = ["0.0", "0.0.1"] # This keeps a record of all versions that the code is equipped to handle. That way, if there's a new data release but they didn't update their package, it won't try to parse the new data version it isn't equipped to handle.

        data_files = {
            "0.0": [
                "clinical.csv.gz",
                "cnv.tsv.gz",
                "definitions.txt",
                "phosphoproteomics.txt.gz",
                "proteomics.txt.gz",
                "somatic_38.maf.gz",
                "transcriptomics.tsv.gz",
                "treatment.csv.gz"],
            "0.0.1": [
                "clinical.csv.gz",
                "cnv.tsv.gz",
                "definitions.txt",
                "Ovary_One_Year_Clinical_Data_20160927.xls",
                "phosphoproteomics.txt.gz",
                "proteomics.txt.gz",
                "somatic_38.maf.gz",
                "transcriptomics.tsv.gz",
                "treatment.csv.gz"],
        }

        super().__init__(cancer_type="ovarian", version=version, valid_versions=valid_versions, data_files=data_files, no_internet=no_internet)

        # Load the data files into dataframes in the self._data dict
        loading_msg = f"Loading {self.get_cancer_type()} v{self.version()}"
        for file_path in self._data_files_paths:

            # Print a loading message. We add a dot every time, so the user knows it's not frozen.
            loading_msg = loading_msg + "."
            print(loading_msg, end='\r')

            path_elements = file_path.split(os.sep) # Get a list of all the levels of the path
            file_name = path_elements[-1] # The last element will be the name of the file
            df_name = file_name.split(".")[0] # Our dataframe name will be the first section of file name (i.e. proteomics.txt.gz becomes proteomics)

            if file_name == "clinical.csv.gz" or file_name == "treatment.csv.gz":
                df = pd.read_csv(file_path, sep=",", index_col=0)
                df = df.rename(columns={"Participant_ID":"Patient_ID"})
                df = df.set_index("Patient_ID")
                self._data[df_name] = df #maps dataframe name to dataframe

            elif file_name == "cnv.tsv.gz":
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                self._data["CNV"] = df #maps dataframe name to dataframe

            elif file_name == "definitions.txt":
                with open(file_path, "r", errors="ignore") as definitions_file:
                    for line in definitions_file.readlines():
                        line = line.strip()
                        line = line.split("\t")
                        term = line[0]
                        definition = line[1]
                        self._definitions[term] = definition

            elif file_name == "phosphoproteomics.txt.gz" or file_name == "proteomics.txt.gz":
                df = pd.read_csv(file_path, sep='\t')
                   
                if file_name == "proteomics.txt.gz":
                    df = df[df["hgnc_symbol"].notnull()] # Drops all nan values in hgnc_symbol column

                    # Create our column multiindex
                    df = df.rename(columns={"hgnc_symbol": "Name", "refseq_peptide": "Database_ID"})
                    df = df.set_index(["Name", "Database_ID"])

                elif file_name == "phosphoproteomics.txt.gz":
                    df = df[df["site"].notnull()] # Drops all rows with nan values in site column

                    # Create our column multiindex
                    split_genes = df["site"].str.rsplit("-", n=1, expand=True) # Split the genes from the sites, splitting from the right since some genes have hyphens in their names, but the genes and sites are also separated by hyphens
                    df = df.drop(columns=["hgnc_symbol", "site"]) # hgnc_symbol is a duplicate of split_genes[0], and site is now in split_genes and will be re-inserted differently
                    df = df.assign(Name=split_genes[0], Site=split_genes[1])
                    df["Site"] = df["Site"].str.replace(r"[sty]", r"", regex=True) # Get rid of all lowercase s, t, and y delimeters in the sites
                    df = df.rename(columns={"refseq_peptide": "Database_ID"})
                    df = df.set_index(["Name", "Site", "Peptide", "Database_ID"]) # Turn these columns into a multiindex

                df = df.sort_index()
                df = df.transpose()
                df.index = df.index.where(~df.index.str.startswith('C'), df.index.str[1:]) # Take C prefix off of indices for those samples that have them (tumor samples have C, normal have N)
                df = df.drop(index=df.index[df.index.str.startswith("OV_QC")]) # Drop all OV_QC samples--they're quality control samples not relevant for data analysis
                self._data[df_name] = df

            elif file_name == "somatic_38.maf.gz":
                df = pd.read_csv(file_path, sep = "\t", index_col=0)
                df = df.reset_index()
                split_barcode = df["Tumor_Sample_Barcode"].str.split("_", n = 1, expand = True) # The first part of the barcode is the patient id, which we need to make a Patient_ID column
                df["Tumor_Sample_Barcode"] = split_barcode[0]
                df = df[["Tumor_Sample_Barcode","Hugo_Symbol","Variant_Classification","HGVSp_Short"]] # We only want these columns
                df = df.rename(columns={"Tumor_Sample_Barcode":"Patient_ID","Hugo_Symbol":"Gene","Variant_Classification":"Mutation","HGVSp_Short":"Location"})
                df = df.sort_values(by=["Patient_ID", "Gene"])
                df = df.set_index("Patient_ID")
                self._data['somatic_mutation'] = df

            elif file_name == "transcriptomics.tsv.gz":
                df = pd.read_csv(file_path, sep="\t", index_col=0)
                df = df.sort_index()
                df = df.transpose()
                df = df.sort_index()
                date_cols = ['1-Dec', '1-Sep', '10-Mar', '10-Sep', '11-Sep', '12-Sep', '14-Sep', '15-Sep', '2-Mar', '2-Sep', '3-Mar', '3-Sep', '4-Mar', '4-Sep', '5-Mar', '6-Mar', '6-Sep', '7-Mar', '7-Sep', '8-Mar', '8-Sep', '9-Mar', '9-Sep']
                df = df.drop(columns=date_cols) # Drop all date values until new data is uploaded
                self._data[df_name] = df #maps dataframe name to dataframe

            elif file_name == 'Ovary_One_Year_Clinical_Data_20160927.xls' and self._version == "0.0.1":
                df = pd.read_excel(file_path)

                # Replace redundant values for "not reported" with NaN
                nan_equivalents = ['Not Reported/ Unknown', 'Reported/ Unknown', 'Not Applicable',
                    'na', 'unknown', 'Not Performed', 'Unknown tumor status', 'Unknown', 
                    'Unknown Tumor Status', 'Not specified']

                df = df.replace(nan_equivalents, np.nan)

                # Rename PPID to Patient_ID and set as index
                df = df.rename(columns={'PPID': 'Patient_ID'})
                df = df.set_index("Patient_ID")
                df = df.sort_index()

                self._data["followup"] = df

        print(' ' * len(loading_msg), end='\r') # Erase the loading message
        formatting_msg = "Formatting dataframes..."
        print(formatting_msg, end='\r')

        # Get a union of all dataframes' indices, with duplicates removed
        master_index = unionize_indices(self._data, exclude="followup")

        # Use the master index to reindex the clinical dataframe, so the clinical dataframe has a record of every sample in the dataset. Rows that didn't exist before (such as the rows for normal samples) are filled with NaN
        master_clinical = self._data['clinical'].reindex(master_index)

        # Add a column called Sample_Tumor_Normal to the clinical dataframe indicating whether each sample was a tumor or normal sample. Normal samples have a Patient_ID that begins with 'N'.
        clinical_status_col = generate_sample_status_col(master_clinical, normal_test=lambda sample: sample[0] == 'N')
        master_clinical.insert(0, "Sample_Tumor_Normal", clinical_status_col)

        # Replace the clinical dataframe in the data dictionary with our new and improved version!
        self._data['clinical'] = master_clinical 

        # Edit the format of the Patient_IDs to have normal samples marked the same way as in other datasets. Currently, all the normal samples have an "N" prepended. We're going to erase that and put a ".N" at the end.
        self._data = reformat_normal_patient_ids(self._data, existing_identifier="N", existing_identifier_location="start")

        # Call function from dataframe_tools.py to sort all tables first by sample status, and then by the index
        self._data = sort_all_rows(self._data)

        # Call function from dataframe_tools.py to standardize the names of the index and column axes
        self._data = standardize_axes_names(self._data)

        print(" " * len(formatting_msg), end='\r') # Erase the formatting message

    def how_to_cite(self):
        return super().how_to_cite(cancer_type='high grade serous ovarian cancer', pmid=27372738)